from . import ExternalItem


class FunctionItem(ExternalItem):
    """
    A menu item to call a Python function
    """

    def __init__(self, text, function, args=None, kwargs=None, menu=None, should_exit=False):
        """
        :ivar function: The function to be called
        :ivar list args: An optional list of arguments to be passed to the function
        :ivar dict kwargs: An optional dictionary of keyword arguments to be passed to the function
        :ivar return_value: the value returned by the function, None if it hasn't been called yet.
        """
        super(FunctionItem, self).__init__(text=text, menu=menu, should_exit=should_exit)

        self.function = function

        if args is not None:
            self.args = args
        else:
            self.args = []
        if kwargs is not None:
            self.kwargs = kwargs
        else:
            self.kwargs = {}

        self.return_value = None

    def action(self):
        """
        This class overrides this method
        """
        self.return_value = self.function(*self.args, **self.kwargs)

    def clean_up(self):
        """
        This class overrides this method
        """
        self.menu.resume()

    def get_return(self):
        """
        :return: The return value from the function call
        """
        return self.return_value



import argparse
import logging
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description


class ConnectCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara connect',
            description=strings['connect'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-d', '--directory',
            help='Specify the folder containing the lab.',
        )
        group.add_argument(
            '-v', '--vmachine',
            dest="vmachine",
            action="store_true",
            help='The device has been started with vstart command.',
        )
        self.parser.add_argument(
            '--shell',
            required=False,
            help='Shell that should be used inside the device.'
        )
        self.parser.add_argument(
            '-l', '--logs',
            action="store_true",
            help='Print device startup logs before launching the shell.',
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            help='Name of the device to connect to.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        if args['vmachine']:
            lab_path = "kathara_vlab"
        else:
            lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
            lab_path = utils.get_absolute_path(lab_path)

        logging.debug("Executing `connect` command in path `%s`..." % lab_path)

        lab_hash = utils.generate_urlsafe_hash(lab_path)

        Kathara.get_instance().connect_tty(lab_hash, machine_name=args['machine_name'], shell=args['shell'],
                                           logs=args['logs'])



import logging
import shlex
from functools import partial
from itertools import islice
from multiprocessing.dummy import Pool
from typing import List, Any, Dict, Generator, Optional, Set, Tuple

import docker.models.containers
import progressbar
from docker import DockerClient
from docker.errors import APIError

from .DockerImage import DockerImage
from ... import utils
from ...exceptions import MountDeniedError, MachineAlreadyExistsError
from ...model.Lab import Lab
from ...model.Link import BRIDGE_LINK_NAME
from ...model.Machine import Machine
from ...setting.Setting import Setting

RP_FILTER_NAMESPACE = "net.ipv4.conf.%s.rp_filter"

# Known commands that each container should execute
# Run order: shared.startup, machine.startup and machine.startup_commands
STARTUP_COMMANDS = [
    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Docker inside the container.
    # In this way, they can be overwritten by custom user files.
    "umount /etc/resolv.conf",
    "umount /etc/hosts",

    # Copy the machine folder (if present) from the hostlab directory into the root folder of the container
    # In this way, files are all replaced in the container root folder
    "if [ -d \"/hostlab/{machine_name}\" ]; then "
    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf -); fi",

    # Give proper permissions to /var/www
    "if [ -d \"/var/www\" ]; then "
    "chmod -R 777 /var/www/*; fi",

    # Give proper permissions to Quagga files (if present)
    "if [ -d \"/etc/quagga\" ]; then "
    "chown quagga:quagga /etc/quagga/*",
    "chmod 640 /etc/quagga/*; fi",

    # Give proper permissions to FRR files (if present)
    "if [ -d \"/etc/frr\" ]; then "
    "chown frr:frr /etc/frr/*",
    "chmod 640 /etc/frr/*; fi",

    # If shared.startup file is present
    "if [ -f \"/hostlab/shared.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/shared.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/shared.startup",
    "/hostlab/shared.startup &> /var/log/shared.log; fi",

    # If .startup file is present
    "if [ -f \"/hostlab/{machine_name}.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/{machine_name}.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/{machine_name}.startup",
    "/hostlab/{machine_name}.startup &> /var/log/startup.log; fi",

    # Placeholder for user commands
    "{machine_commands}"
]

SHUTDOWN_COMMANDS = [
    # If machine.shutdown file is present
    "if [ -f \"/hostlab/{machine_name}.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/{machine_name}.shutdown; /hostlab/{machine_name}.shutdown; fi",

    # If shared.shutdown file is present
    "if [ -f \"/hostlab/shared.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/shared.shutdown; /hostlab/shared.shutdown; fi"
]


class DockerMachine(object):
    """The class responsible for deploying Kathara devices as Docker container and interact with them."""
    __slots__ = ['client', 'docker_image']

    def __init__(self, client: DockerClient, docker_image: DockerImage) -> None:
        self.client: DockerClient = client

        self.docker_image: DockerImage = docker_image

    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """Deploy all the lab devices as Docker containers.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_machines (Set[str]): A set containing the name of the devices to deploy.

        Returns:
            None
        """
        # Check and pulling machine images
        lab_images = set(map(lambda x: x.get_image(), lab.machines.values()))
        self.docker_image.check_and_pull_from_list(lab_images)

        shared_mount = lab.general_options['shared_mount'] if 'shared_mount' in lab.general_options \
            else Setting.get_instance().shared_mount

        if shared_mount:
            if Setting.get_instance().remote_url is not None:
                logging.warning("Shared folder cannot be mounted with a remote Docker connection.")
            else:
                lab.create_shared_folder()

        machines = {k: v for (k, v) in lab.machines.items() if k in selected_machines}.items() if selected_machines \
            else lab.machines.items()

        progress_bar = None
        if utils.CLI_ENV:
            progress_bar = progressbar.ProgressBar(
                widgets=['Deploying devices... ', progressbar.Bar(),
                         ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                redirect_stdout=True,
                max_value=len(machines)
            )

        # Deploy all lab machines.
        # If there is no lab.dep file, machines can be deployed using multithreading.
        # If not, they're started sequentially
        if not lab.has_dependencies:
            pool_size = utils.get_pool_size()
            machines_pool = Pool(pool_size)

            items = utils.chunk_list(machines, pool_size)

            for chunk in items:
                machines_pool.map(func=partial(self._deploy_and_start_machine, progress_bar), iterable=chunk)
        else:
            for item in machines:
                self._deploy_and_start_machine(progress_bar, item)

        if utils.CLI_ENV:
            progress_bar.finish()

    def _deploy_and_start_machine(self, progress_bar: Optional[progressbar.ProgressBar],
                                  machine_item: Tuple[str, Machine]) -> None:
        """Deploy and start a Docker container from the device contained in machine_item.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            machine_item (Tuple[str, Machine]): A tuple composed by the name of the device and a device object

        Returns:
            None
        """
        (_, machine) = machine_item

        self.create(machine)
        self.start(machine)

        if progress_bar is not None:
            progress_bar += 1

    def create(self, machine: Machine) -> None:
        """Create a Docker container representing the device and assign it to machine.api_object.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            None
        """
        logging.debug("Creating device `%s`..." % machine.name)

        machines = self.get_machines_api_objects_by_filters(machine_name=machine.name, lab_hash=machine.lab.hash,
                                                            user=utils.get_current_user_name())
        if machines:
            raise MachineAlreadyExistsError("Device with name `%s` already exists." % machine.name)

        image = machine.get_image()
        memory = machine.get_mem()
        cpus = machine.get_cpu(multiplier=1000000000)

        ports_info = machine.get_ports()
        ports = None
        if ports_info:
            ports = {}
            for (host_port, protocol), guest_port in ports_info.items():
                ports['%d/%s' % (guest_port, protocol)] = host_port

        # Get the general options into a local variable (just to avoid accessing the lab object every time)
        options = machine.lab.general_options

        # If bridged is required in command line but not defined in machine meta, add it.
        if "bridged" in options and not machine.meta['bridged']:
            machine.add_meta("bridged", True)

        # If any exec command is passed in command line, add it.
        if "exec" in options:
            machine.add_meta("exec", options["exec"])

        # Get the first network object, if defined.
        # This should be used in container create function
        first_network = None
        if machine.interfaces:
            first_network = machine.interfaces[0].api_object

        # If no interfaces are declared in machine, but bridged mode is required, get bridge as first link.
        # Flag that bridged is already connected (because there's another check in `start`).
        if first_network is None and machine.meta['bridged']:
            first_network = machine.lab.get_or_new_link(BRIDGE_LINK_NAME).api_object
            machine.add_meta("bridge_connected", True)

        # Sysctl params to pass to the container creation
        sysctl_parameters = {RP_FILTER_NAMESPACE % x: 0 for x in ["all", "default", "lo"]}

        if first_network:
            sysctl_parameters[RP_FILTER_NAMESPACE % "eth0"] = 0

        sysctl_parameters["net.ipv4.ip_forward"] = 1
        sysctl_parameters["net.ipv4.icmp_ratelimit"] = 0

        if machine.is_ipv6_enabled():
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 1
            sysctl_parameters["net.ipv6.icmp.ratelimit"] = 0
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 0
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 0

        # Merge machine sysctls
        sysctl_parameters = {**sysctl_parameters, **machine.meta['sysctls']}

        volumes = {}

        shared_mount = options['shared_mount'] if 'shared_mount' in options else Setting.get_instance().shared_mount
        if shared_mount and machine.lab.shared_folder:
            volumes[machine.lab.shared_folder] = {'bind': '/shared', 'mode': 'rw'}

        # Mount the host home only if specified in settings.
        hosthome_mount = options['hosthome_mount'] if 'hosthome_mount' in options else \
            Setting.get_instance().hosthome_mount
        if hosthome_mount and Setting.get_instance().remote_url is None:
            volumes[utils.get_current_user_home()] = {'bind': '/hosthome', 'mode': 'rw'}

        privileged = options['privileged_machines'] if 'privileged_machines' in options else False
        if Setting.get_instance().remote_url is not None and privileged:
            privileged = False
            logging.warning("Privileged flag is ignored with a remote Docker connection.")

        container_name = self.get_container_name(machine.name, machine.lab.hash)

        try:
            machine_container = self.client.containers.create(image=image,
                                                              name=container_name,
                                                              hostname=machine.name,
                                                              cap_add=machine.capabilities if not privileged else None,
                                                              privileged=privileged,
                                                              network=first_network.name if first_network else None,
                                                              network_mode="bridge" if first_network else "none",
                                                              sysctls=sysctl_parameters,
                                                              mem_limit=memory,
                                                              nano_cpus=cpus,
                                                              ports=ports,
                                                              tty=True,
                                                              stdin_open=True,
                                                              detach=True,
                                                              volumes=volumes,
                                                              labels={"name": machine.name,
                                                                      "lab_hash": machine.lab.hash,
                                                                      "user": utils.get_current_user_name(),
                                                                      "app": "kathara",
                                                                      "shell": machine.meta["shell"]
                                                                      if "shell" in machine.meta
                                                                      else Setting.get_instance().device_shell
                                                                      }
                                                              )
        except APIError as e:
            raise e

        # Pack machine files into a tar.gz and extract its content inside `/`
        tar_data = machine.pack_data()
        if tar_data:
            self.copy_files(machine_container, "/", tar_data)

        machine.api_object = machine_container

    def update(self, machine: Machine) -> None:
        """Update the Docker container representing the machine.

        Create a new Docker network for each link contained in
        machine.interfaces that is not already attached to the container.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            None
        """
        machines = self.get_machines_api_objects_by_filters(machine_name=machine.name, lab_hash=machine.lab.hash,
                                                            user=utils.get_current_user_name())

        if not machines:
            raise Exception("Device `%s` not found." % machine.name)

        machine.api_object = machines.pop()
        attached_networks = machine.api_object.attrs["NetworkSettings"]["Networks"]

        # Connect the container to its new networks
        for (_, machine_link) in machine.interfaces.items():
            if machine_link.api_object.name not in attached_networks:
                machine_link.api_object.connect(machine.api_object)

    @staticmethod
    def start(machine: Machine) -> None:
        """Start the Docker container representing the device.

        Connect the container to the networks, run the startup commands and open a terminal (if requested).

        Args:
           machine (Kathara.model.Machine.Machine): A Kathara device.
        """
        logging.debug("Starting device `%s`..." % machine.name)

        try:
            machine.api_object.start()
        except APIError as e:
            if e.response.status_code == 500 and e.explanation.startswith('Mounts denied'):
                raise MountDeniedError("Host drive is not shared with Docker.")
            else:
                raise e

        # Connect the container to its networks (starting from the second, the first is already connected in `create`)
        # This should be done after the container start because Docker causes a non-deterministic order when attaching
        # networks before container startup.
        for (iface_num, machine_link) in islice(machine.interfaces.items(), 1, None):
            logging.debug("Connecting device `%s` to collision domain `%s` on interface %d..." % (machine.name,
                                                                                                  machine_link.name,
                                                                                                  iface_num
                                                                                                  )
                          )

            machine_link.api_object.connect(machine.api_object)

        # Bridged connection required but not added in `deploy` method.
        if "bridge_connected" not in machine.meta and machine.meta['bridged']:
            bridge_link = machine.lab.get_or_new_link(BRIDGE_LINK_NAME).api_object
            bridge_link.connect(machine.api_object)

        # Append executed machine startup commands inside the /var/log/startup.log file
        if machine.startup_commands:
            new_commands = []
            for command in machine.startup_commands:
                new_commands.append("echo \"++ %s\" &>> /var/log/startup.log" % command)
                new_commands.append(command)
            machine.startup_commands = new_commands

        # Build the final startup commands string
        startup_commands_string = "; ".join(STARTUP_COMMANDS).format(
            machine_name=machine.name,
            machine_commands="; ".join(machine.startup_commands)
        )

        # Execute the startup commands inside the container (without privileged flag so basic permissions are used)
        machine.api_object.exec_run(cmd=[Setting.get_instance().device_shell, '-c', startup_commands_string],
                                    stdout=False,
                                    stderr=False,
                                    privileged=False,
                                    detach=True
                                    )

        if Setting.get_instance().open_terminals:
            for i in range(0, machine.get_num_terms()):
                machine.connect(Setting.get_instance().terminal)

    def undeploy(self, lab_hash: str, selected_machines: Set[str] = None) -> None:
        """Undeploy the devices contained in the network scenario defined by the lab_hash.

        If a set of selected_machines is specified, undeploy only the specified devices.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        machines = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, user=utils.get_current_user_name())

        if selected_machines is not None and len(selected_machines) > 0:
            machines = [item for item in machines if item.labels["name"] in selected_machines]

        if len(machines) > 0:
            pool_size = utils.get_pool_size()
            machines_pool = Pool(pool_size)

            items = utils.chunk_list(machines, pool_size)

            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deleting devices... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(machines)
                )

            for chunk in items:
                machines_pool.map(func=partial(self._undeploy_machine, progress_bar),
                                  iterable=chunk
                                  )
            if utils.CLI_ENV:
                progress_bar.finish()

    def wipe(self, user: str = None) -> None:
        """Undeploy all the running devices of the specified user. If user is None, it undeploy all the running devices.

        Args:
            user (str): The name of a current user on the host.

        Returns:
            None
        """
        machines = self.get_machines_api_objects_by_filters(user=user)

        pool_size = utils.get_pool_size()
        machines_pool = Pool(pool_size)

        items = utils.chunk_list(machines, pool_size)

        for chunk in items:
            machines_pool.map(func=partial(self._undeploy_machine, None), iterable=chunk)

    def _undeploy_machine(self, progress_bar: Optional[progressbar.ProgressBar],
                          machine_api_object: docker.models.containers.Container) -> None:
        """Undeploy a Docker container.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            machine_api_object (docker.models.containers.Container): The Docker container to undeploy.

        Returns:
            None
        """
        self._delete_machine(machine_api_object)

        if progress_bar is not None:
            progress_bar += 1

    def connect(self, lab_hash: str, machine_name: str, user: str = None, shell: str = None,
                logs: bool = False) -> None:
        """Open a stream to the Docker container specified by machine_name using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device to connect.
            user (str): The name of a current user on the host.
            shell (str): The path to the desired shell.
            logs (bool): If True, print the logs of the startup command.

        Returns:
            None
        """
        container = self.get_machine_api_object(lab_hash=lab_hash, machine_name=machine_name, user=user)

        if not shell:
            shell = shlex.split(container.labels['shell'])
        else:
            shell = shlex.split(shell) if type(shell) == str else shell

        logging.debug("Connect to device `%s` with shell: %s" % (machine_name, shell))

        if logs and Setting.get_instance().print_startup_log:
            exec_output = self.exec(lab_hash,
                                    machine_name,
                                    user=user,
                                    command="cat /var/log/shared.log /var/log/startup.log",
                                    tty=False
                                    )

            startup_output = ""
            try:
                while True:
                    (stdout, _) = next(exec_output)
                    startup_output += stdout.decode('utf-8') if stdout else ""
            except StopIteration:
                pass

            if startup_output:
                print("--- Startup Commands Log\n")
                print(startup_output)
                print("--- End Startup Commands Log\n")

        resp = self.client.api.exec_create(container.id,
                                           shell,
                                           stdout=True,
                                           stderr=True,
                                           stdin=True,
                                           tty=True,
                                           privileged=False
                                           )

        exec_output = self.client.api.exec_start(resp['Id'],
                                                 tty=True,
                                                 socket=True
                                                 )

        def tty_connect():
            from .terminal.DockerTTYTerminal import DockerTTYTerminal
            DockerTTYTerminal(exec_output, self.client, resp['Id']).start()

        def cmd_connect():
            from .terminal.DockerNPipeTerminal import DockerNPipeTerminal
            DockerNPipeTerminal(exec_output, self.client, resp['Id']).start()

        utils.exec_by_platform(tty_connect, cmd_connect, tty_connect)

    def exec(self, lab_hash: str, machine_name: str, command: str, user: str = None,
             tty: bool = True) -> Generator[Tuple[bytes, bytes], None, None]:
        """Execute the command on the Docker container specified by the lab_hash and the machine_name.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device.
            user (str): The name of a current user on the host.
            command (str): The command to execute.
            tty (bool): If True, open a new tty.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        logging.debug("Executing command `%s` to device with name: %s" % (command, machine_name))

        container = self.get_machine_api_object(lab_hash=lab_hash, machine_name=machine_name, user=user)

        exec_result = container.exec_run(cmd=command,
                                         stdout=True,
                                         stderr=True,
                                         tty=tty,
                                         privileged=False,
                                         stream=True,
                                         demux=True,
                                         detach=False
                                         )

        return exec_result.output

    @staticmethod
    def copy_files(machine_api_object: docker.models.containers.Container, path: str, tar_data: bytes) -> None:
        """Copy the files contained in tar_data in the Docker container path specified by the machine_api_object.

        Args:
            machine_api_object (docker.models.containers.Container): A Docker container.
            path (str): The path of the container where copy the tar_data.
            tar_data (bytes): The data to copy in the container.

        Returns:
            None
        """
        machine_api_object.put_archive(path, tar_data)

    def get_machines_api_objects_by_filters(self, lab_hash: str = None, machine_name: str = None, user: str = None) -> \
            List[docker.models.containers.Container]:
        """Return the Docker containers objects specified by lab_hash, machine_name and user.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the devices in the scenario.
            machine_name (str): The name of a device. If specified, return the specified container of the scenario.
            user (str): The name of a user on the host. If specified, return only the containers of the user.

        Returns:
            List[docker.models.containers.Container]: A list of Docker containers.
        """
        filters = {"label": ["app=kathara"]}
        if user:
            filters["label"].append("user=%s" % user)
        if lab_hash:
            filters["label"].append("lab_hash=%s" % lab_hash)
        if machine_name:
            filters["label"].append("name=%s" % machine_name)

        return self.client.containers.list(all=True, filters=filters)

    def get_machine_api_object(self, lab_hash: str, machine_name: str,
                               user: str = None) -> docker.models.containers.Container:
        """Return the Docker container object specified by lab_hash and machine_name.

        Args:
            lab_hash (str): The hash of a network scenario.
            machine_name (str): The name of a device.
            user (str): The name of a current user on the host.

        Returns:
            docker.models.containers.Container: A Docker container.
        """
        logging.debug("Searching container `%s` with lab hash `%s` for user `%s`" % (machine_name, lab_hash, user))

        containers = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name, user=user)
        logging.debug("Found containers: %s" % str(containers))

        if len(containers) != 1:
            raise Exception("Error getting the device `%s` inside the lab." % machine_name)
        else:
            return containers[0]

    def get_machine_info(self, machine_name: str, lab_hash: str = None, user: str = None) -> List[Dict[str, Any]]:
        """Return a list of dicts containing the devices info.

        Args:
            machine_name (str): The name of a device
            lab_hash (str): The hash of a network scenario. If specified, search only the devices in the corresponding
            scenario.
            user (str): The name of a user on the host. If specified, search only the containers of the user.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        machines_api_objects = self.get_machines_api_objects_by_filters(machine_name=machine_name, lab_hash=lab_hash,
                                                                        user=user)
        if not machines_api_objects:
            raise Exception("The specified device `%s` is not running." % machine_name)

        all_stats = []
        for machine_api_object in machines_api_objects:
            machine_stats = machine_api_object.stats(stream=False)
            all_stats.append(self._get_stats_by_machine(machine_api_object, machine_stats))

        return all_stats

    def get_machines_info(self, lab_hash: str = None, machine_name: str = None, user: str = None) -> \
            Generator[Dict[str, Any], None, None]:
        """Return a generator containing the info of the specified device.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the devices in the scenario.
            machine_name (str): The name of a device. If specified, return the specified container of the scenario.
            user (str): The name of a user on the host. If specified, return only the containers of the user.

        Returns:
            Generator[Dict[str, Any], None, None]: A generator containing the info of the specified device.
        """
        machines = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name, user=user)
        if not machines:
            if not lab_hash:
                raise Exception("No devices running.")
            else:
                raise Exception("Lab is not started.")

        machines = sorted(machines, key=lambda x: x.name)

        machine_streams = {}

        for machine in machines:
            machine_streams[machine] = machine.stats(stream=True, decode=True)

        while True:
            machines_data = []

            for (machine, machine_stats) in machine_streams.items():
                try:
                    result = next(machine_stats)
                except StopIteration:
                    continue

                machines_data.append(self._get_stats_by_machine(machine, result))

            yield machines_data

    def _get_stats_by_machine(self, machine_api_object: docker.models.containers.Container,
                              machine_stats: Dict[str, Any]) -> Dict[str, Any]:
        """Return the stats of the specified Docker container.

        Args:
            machine_api_object (docker.models.containers.Container): A Docker container.
            machine_stats (Dict[str, Any]): A dict containing the stats of the container from Docker.

        Returns:
            Dict[str, Any]: A dict containing formatted Kathara device stats.

        """
        stats = self._get_aggregate_machine_info(machine_stats)

        return {
            "real_lab_hash": machine_api_object.labels['lab_hash'],
            "name": machine_api_object.labels['name'],
            "real_name": machine_api_object.name,
            "user": machine_api_object.labels['user'],
            "status": machine_api_object.status,
            "image": machine_api_object.image.tags[0],
            "pids": machine_stats['pids_stats']['current'] if 'current' in machine_stats['pids_stats'] else 0,
            "cpu_usage": stats['cpu_usage'],
            "mem_usage": stats['mem_usage'],
            "mem_percent": stats['mem_percent'],
            "net_usage": stats['net_usage']
        }

    @staticmethod
    def _get_aggregate_machine_info(stats: Dict[str, Any]) -> Dict[str, str]:
        """Return a dict containing the aggregate Kathara device info.

        Args:
             stats (Dict[str, Any]): A dict containing the stats of the container from Docker.

        Returns:
            Dict[str, Any]: A dict containing formatted Docker container stats.
        """
        network_stats = stats["networks"] if "networks" in stats else {}

        return {
            "cpu_usage": "{0:.2f}%".format(stats["cpu_stats"]["cpu_usage"]["total_usage"] /
                                           stats["cpu_stats"]["system_cpu_usage"]
                                           ) if "system_cpu_usage" in stats["cpu_stats"] else "-",
            "mem_usage": utils.human_readable_bytes(stats["memory_stats"]["usage"]) + " / " +
                         utils.human_readable_bytes(stats["memory_stats"]["limit"])
            if "usage" in stats["memory_stats"] else "- / -",
            "mem_percent": "{0:.2f}%".format((stats["memory_stats"]["usage"] / stats["memory_stats"]["limit"]) * 100)
            if "usage" in stats["memory_stats"] else "-",
            "net_usage": utils.human_readable_bytes(sum([net_stats["rx_bytes"]
                                                         for (_, net_stats) in network_stats.items()])
                                                    ) + " / " +
                         utils.human_readable_bytes(sum([net_stats["tx_bytes"]
                                                         for (_, net_stats) in network_stats.items()])
                                                    )
        }

    @staticmethod
    def get_container_name(name: str, lab_hash: str) -> str:
        """Return the name of a Docker container.

        Args:
            name (str): The name of a Kathara device.
            lab_hash (str): The hash of a running scenario.

        Returns:
            str: The name of the Docker container in the format "|dev_prefix|_|username_prefix|_|name|_|lab_hash|".
        """
        lab_hash = lab_hash if "_%s" % lab_hash else ""
        return "%s_%s_%s_%s" % (Setting.get_instance().device_prefix, utils.get_current_user_name(), name, lab_hash)

    @staticmethod
    def _delete_machine(machine: docker.models.containers.Container) -> None:
        """Remove a running Docker container.

        Args:
            machine (docker.models.containers.Container): The Docker container to remove.

        Returns:
            None
        """
        # Build the shutdown command string
        shutdown_commands_string = "; ".join(SHUTDOWN_COMMANDS).format(machine_name=machine.labels["name"])

        # Execute the shutdown commands inside the container (only if it's running)
        if machine.status == "running":
            machine.exec_run(cmd=[Setting.get_instance().device_shell, '-c', shutdown_commands_string],
                             stdout=False,
                             stderr=False,
                             privileged=True,
                             detach=True
                             )

        machine.remove(force=True)



import logging
from typing import Union, List, Set

import docker.models.images
from docker import DockerClient
from docker.errors import APIError

from ... import utils


class DockerImage(object):
    """Class responsible for interacting with Docker Images."""
    __slots__ = ['client']

    def __init__(self, client: DockerClient) -> None:
        self.client: DockerClient = client

    def get_local(self, image_name: str) -> docker.models.images.Image:
        """Return the specified Docker Image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            docker.models.images.Image: A Docker Image
        """
        return self.client.images.get(image_name)

    def get_remote(self, image_name: str) -> docker.models.images.RegistryData:
        """Gets the registry data for an image.

        Args:
            image_name (str): The name of the image.

        Returns:
            docker.models.images.RegistryData: The data object.

        Raises:
            `docker.errors.APIError`: If the server returns an error.
        """
        return self.client.images.get_registry_data(image_name)

    def pull(self, image_name: str) -> docker.models.images.Image:
        """Pull and return the specified Docker Image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            docker.models.images.Image: A Docker Image
        """
        # If no tag or sha key is specified, we add "latest"
        if (':' or '@') not in image_name:
            image_name = "%s:latest" % image_name

        logging.info("Pulling image `%s`... This may take a while." % image_name)
        return self.client.images.pull(image_name)

    def check_for_updates(self, image_name: str) -> None:
        """Update the specified image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            None
        """
        logging.debug("Checking updates for %s..." % image_name)

        if '@' in image_name:
            logging.debug('No need to check image digest of %s' % image_name)
            return

        local_image_info = self.get_local(image_name)
        # Image has been built locally, so there's nothing to compare.
        local_repo_digests = local_image_info.attrs["RepoDigests"]
        if not local_repo_digests:
            logging.debug("Image %s is build locally" % image_name)
            return

        remote_image_info = self.get_remote(image_name).attrs['Descriptor']
        local_repo_digest = local_repo_digests[0]
        remote_image_digest = remote_image_info["digest"]

        # Format is image_name@sha256, so we strip the first part.
        (_, local_image_digest) = local_repo_digest.split("@")
        # We only need to update tagged images, not the ones with digests.
        if remote_image_digest != local_image_digest:
            if utils.CLI_ENV:
                utils.confirmation_prompt("A new version of image `%s` has been found on Docker Hub. "
                                          "Do you want to pull it?" % image_name,
                                          lambda: self.pull(image_name),
                                          lambda: None
                                          )
            else:
                self.pull(image_name)

    def check(self, image_name: str) -> None:
        """Check the existence of the specified image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            None
        """
        self._check_and_pull(image_name, pull=False)

    def check_and_pull_from_list(self, images: Union[List[str], Set[str]]) -> None:
        """Check and pull a list of specified images.

        Args:
            images (Union[List[str], Set[str]]): A list of Docker images name to pull.

        Returns:
            None
        """
        for image in images:
            self._check_and_pull(image)

    def _check_and_pull(self, image_name: str, pull: bool = True) -> None:
        """Check and pull of the specified image.

        Args:
            image_name (str): The name of a Docker Image.
            pull (bool): If True, pull the image from Docker Hub.

        Returns:
            None
        """
        try:
            # Tries to get the image from the local Docker repository.
            self.get_local(image_name)
            try:
                if pull:
                    self.check_for_updates(image_name)
            except APIError:
                logging.debug("Cannot check updates, skipping...")
        except APIError:
            # If not found, tries on Docker Hub.
            try:
                # If the image exists on Docker Hub, pulls it.
                self.get_remote(image_name)
                if pull:
                    self.pull(image_name)
            except APIError as e:
                if e.response.status_code == 500 and 'dial tcp' in e.explanation:
                    raise ConnectionError("Image `%s` not found in local and no Internet connection." % image_name)
                else:
                    raise Exception("Image `%s` does not exists either in local or on Docker Hub." % image_name)



import io
import json
from datetime import datetime
from typing import Set, Dict, Generator, Any, List, Tuple

from kubernetes import client
from kubernetes.client.rest import ApiException
from terminaltables import DoubleTable

from .KubernetesConfig import KubernetesConfig
from .KubernetesLink import KubernetesLink
from .KubernetesMachine import KubernetesMachine
from .KubernetesNamespace import KubernetesNamespace
from ... import utils
from ...exceptions import NotSupportedError
from ...foundation.manager.IManager import IManager
from ...model.Lab import Lab
from ...model.Machine import Machine
from ...utils import pack_files_for_tar


class KubernetesManager(IManager):
    """Class responsible for interacting with Kubernetes API."""

    __slots__ = ['k8s_namespace', 'k8s_machine', 'k8s_link']

    def __init__(self) -> None:
        KubernetesConfig.load_kube_config()

        self.k8s_namespace: KubernetesNamespace = KubernetesNamespace()
        self.k8s_machine: KubernetesMachine = KubernetesMachine(self.k8s_namespace)
        self.k8s_link: KubernetesLink = KubernetesLink()

    def deploy_lab(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Set[str]): If not None, deploy only the specified devices.

        Returns:
            None
        """
        # Kubernetes needs only lowercase letters for resources.
        # We force the hash to be lowercase
        lab.hash = lab.hash.lower()

        selected_links = None
        if selected_machines:
            selected_links = lab.get_links_from_machines(selected_machines)

        self.k8s_namespace.create(lab)
        try:
            self.k8s_link.deploy_links(lab, selected_links=selected_links)

            self.k8s_machine.deploy_machines(lab, selected_machines=selected_machines)
        except ApiException as e:
            if e.status == 403 and 'Forbidden' in e.reason:
                raise Exception("Previous lab execution is still terminating. Please wait.")
            else:
                raise e

    def update_lab(self, lab: Lab) -> None:
        """Update a running network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.

        Raises:
            NotSupportedError: "Unable to update a running lab."
        """
        raise NotSupportedError("Unable to update a running lab.")

    def undeploy_lab(self, lab_hash: str, selected_machines: Set[str] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        lab_hash = lab_hash.lower()

        # When only some machines should be undeployed, special checks are required.
        if selected_machines:
            # Get all current deployed networks and save only their name
            networks = self.k8s_link.get_links_api_objects_by_filters(lab_hash=lab_hash)
            all_networks = set([network["metadata"]["name"] for network in networks])

            # Get all current running machines (not Terminating)
            running_machines = [machine for machine in
                                self.k8s_machine.get_machines_api_objects_by_filters(lab_hash=lab_hash)
                                if 'Terminating' not in machine.status.phase
                                ]

            # From machines, save a set with all the attached networks (still needed)
            running_networks = set()
            for machine in running_machines:
                network_annotation = json.loads(machine.metadata.annotations["k8s.v1.cni.cncf.io/networks"])
                networks = [net['name'] for net in network_annotation]

                if machine.metadata.labels["name"] not in selected_machines:
                    running_networks.update(networks)

            # Difference between all networks and attached networks are the ones to delete
            networks_to_delete = all_networks - running_networks

            # Save only the fancy name of the machines
            running_machines = set([machine.metadata.labels["name"] for machine in running_machines])
        else:
            networks_to_delete = None
            running_machines = set()

        self.k8s_machine.undeploy(lab_hash, selected_machines=selected_machines)
        self.k8s_link.undeploy(lab_hash, networks_to_delete=networks_to_delete)

        # If no machines are selected or there are no running machines, undeploy the namespace
        if not selected_machines or len(running_machines - selected_machines) <= 0:
            self.k8s_namespace.undeploy(lab_hash=lab_hash)

    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        Args:
        all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
           running network scenarios of all users.

        Returns:
            None
        """
        if all_users:
            raise NotSupportedError("Cannot use `--all` flag.")

        self.k8s_machine.wipe()
        self.k8s_link.wipe()

        self.k8s_namespace.wipe()

    def connect_tty(self, lab_hash: str, machine_name: str, shell: str = None, logs: bool = False) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.

        Returns:
            None
        """
        lab_hash = lab_hash.lower()

        self.k8s_machine.connect(lab_hash=lab_hash,
                                 machine_name=machine_name,
                                 shell=shell,
                                 logs=logs
                                 )

    def exec(self, lab_hash: str, machine_name: str, command: str) -> Generator[Tuple[bytes, bytes], None, None]:
        """Exec a command on a device in a running network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            command (str): The command to exec on the device

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        lab_hash = lab_hash.lower()

        return self.k8s_machine.exec(lab_hash, machine_name, command, stderr=True, tty=False, is_stream=True)

    def copy_files(self, machine: Machine, guest_to_host: Dict[str, io.IOBase]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running device object. It must have the api_object field populated.
            guest_to_host (Dict[str, io.IOBase]): A dict containing the device path as key and
                fileobj to copy in path as value.

        Returns:
            None
        """
        tar_data = pack_files_for_tar(guest_to_host)

        self.k8s_machine.copy_files(machine.api_object,
                                    path="/",
                                    tar_data=tar_data
                                    )

    def get_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> \
            Generator[Dict[str, Any], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, Any], None, None]: A generator containing dicts containing device names as keys and
              their info as values.
        """
        if lab_hash:
            lab_hash = lab_hash.lower()

        lab_info = self.k8s_machine.get_machines_info(lab_hash=lab_hash, machine_name=machine_name)

        return lab_info

    def get_formatted_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> str:
        """Return a formatted string with the information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
             str: String containing devices info
        """
        if all_users:
            raise NotSupportedError("Cannot use `--all` flag on Megalos.")

        table_header = ["LAB HASH", "DEVICE NAME", "STATUS", "ASSIGNED NODE"]
        stats_table = DoubleTable([])
        stats_table.inner_row_border = True

        lab_info = self.get_lab_info(lab_hash=lab_hash, machine_name=machine_name)

        while True:
            machines_data = [
                table_header
            ]

            try:
                result = next(lab_info)
            except StopIteration:
                return

            if not result:
                return

            for machine_stats in result:
                machines_data.append([machine_stats["real_lab_hash"],
                                      machine_stats["name"],
                                      machine_stats["status"],
                                      machine_stats["assigned_node"]
                                      ])

            stats_table.table_data = machines_data

            yield "TIMESTAMP: %s" % datetime.now() + "\n\n" + stats_table.table

    def get_machine_api_object(self, lab_hash: str, machine_name: str) -> client.V1Pod:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            lab_hash (str): The hash of the network scenario.
            machine_name (str): The name of the device.

        Returns:
            client.V1Pod: A Kubernetes PoD.
        """
        lab_hash = lab_hash.lower()
        return self.k8s_machine.get_machine_api_object(lab_hash, machine_name)

    def get_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) \
            -> List[Dict[str, Any]]:
        """Return information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        if lab_hash:
            lab_hash = lab_hash.lower()

        machine_stats = self.k8s_machine.get_machine_info(machine_name=machine_name, lab_hash=lab_hash)

        return machine_stats

    def get_formatted_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) -> str:
        """Return formatted information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            str: The formatted devices properties.
        """
        if all_users:
            raise NotSupportedError("Cannot use `--all` flag.")

        machines_stats = self.get_machine_info(machine_name, lab_hash=lab_hash)

        machines_info = []

        for machine_stats in machines_stats:
            machine_info = utils.format_headers("Device information") + "\n"
            machine_info += "Lab Hash: %s\n" % machine_stats["real_lab_hash"]
            machine_info += "Device Name: %s\n" % machine_stats["name"]
            machine_info += "Real Device Name: %s\n" % machine_stats["real_name"]
            machine_info += "Status: %s\n" % machine_stats["status"]
            machine_info += "Image: %s\n" % machine_stats["image"]
            machine_info += "Assigned Node: %s\n" % machine_stats["assigned_node"]
            machine_info += utils.format_headers()

            machines_info.append(machine_info)

        return "\n\n".join(machines_info)

    def check_image(self, image_name: str) -> None:
        """Useless. The Check of the image is delegated to Kubernetes.

        Args:
            image_name (str): The name of the image

        Returns:
            None
        """
        # Delegate the image check to Kubernetes
        return

    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        return client.VersionApi().get_code().git_version

    @staticmethod
    def get_formatted_manager_name() -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        return "Kubernetes (Megalos)"



from abc import ABC, abstractmethod
from typing import List, Dict, Any

from ..CliArgs import CliArgs


class Command(ABC):
    __slots__ = ['parser']

    def __init__(self) -> None:
        pass

    @abstractmethod
    def run(self, current_path: str, argv: List[str]) -> Any:
        raise NotImplementedError("You must implement `run` method.")

    def parse_args(self, argv: List[str]) -> None:
        args = self.parser.parse_args(argv)
        CliArgs.get_instance().args = vars(args)

    @staticmethod
    def get_args() -> Dict[str, Any]:
        return CliArgs.get_instance().args



from ..setting.Setting import Setting

from ..trdparty.consolemenu.validators.base import BaseValidator


class TerminalValidator(BaseValidator):
    def __init__(self) -> None:
        super(TerminalValidator, self).__init__()

    def validate(self, input_string: str) -> bool:
        try:
            return Setting.get_instance().check_terminal(input_string)
        except Exception:
            return False



import logging
import os
from typing import Optional


class Networking(object):
    """
    Class responsible for managing ExternalLink objects attaching Kathara collision domain to host interfaces.
    """

    @staticmethod
    def get_or_new_interface(full_interface_name: str, vlan_interface_name: str, vlan_id: Optional[int] = None) -> int:
        """
        Get or create an interface on the host. Return the link index.

        Args:
            full_interface_name (str): The name of the network interface of the host.
            vlan_interface_name (str): The name of the corresponding VLAN interface (MAX 15 chars).
            vlan_id (Optional[int]): The VLAN ID. If specified get or create a VLAN interface.

        Returns:
            int: The link index.
        """
        # disable pyroute2 logging to avoid warning about project structure changes
        logging.getLogger('pyroute2').disabled = True
        from pyroute2 import IPRoute
        ip = IPRoute()
        logging.getLogger('pyroute2').disabled = False

        logging.debug("Searching for interface `%s`..." % full_interface_name)

        # Search the interface
        interface_indexes = ip.link_lookup(ifname=full_interface_name)
        # If not present, raise an error
        if not interface_indexes:
            raise Exception("Interface `%s` not found." % full_interface_name)

        interface_index = interface_indexes[0]
        logging.debug("Interface found with ID = %d" % interface_index)

        if vlan_id:
            full_vlan_iface_name = "%s.%s" % (vlan_interface_name, vlan_id)
            logging.debug("VLAN Interface required... Creating `%s`..." % full_vlan_iface_name)

            # Search the VLAN interface
            vlan_link_indexes = ip.link_lookup(ifname=full_vlan_iface_name)

            if not vlan_link_indexes:
                # A VLAN interface should be created before attaching it to bridge.
                ip.link(
                    "add",
                    ifname=full_vlan_iface_name,
                    kind="vlan",
                    link=interface_index,
                    vlan_id=vlan_id
                )

                # Set the new interface up
                ip.link(
                    "set",
                    index=interface_index,
                    state="up"
                )

                logging.debug("Interface `%s` set UP." % full_vlan_iface_name)

                # Refresh the VLAN interface information
                vlan_link_indexes = ip.link_lookup(ifname=full_vlan_iface_name)
                interface_index = vlan_link_indexes[0]

        ip.close()

        return interface_index

    @staticmethod
    def attach_interface_to_bridge(interface_index: int, bridge_name: str) -> None:
        """
        Attach an interface to the bridge.

        Args:
            interface_index (int): The interface index of the interface to attach.
            bridge_name (str):  The name of the bridge to attach to the interface.

        Returns:
            None
        """
        from pyroute2 import IPRoute
        ip = IPRoute()

        logging.debug("Attaching interface ID = %d to bridge `%s`..." % (interface_index, bridge_name))
        bridge_index = ip.link_lookup(ifname=bridge_name)[0]

        ip.link(
            "set",
            index=interface_index,
            master=bridge_index,
            state="up"
        )

        logging.debug("Interface ID = %d attached to bridge %s." % (interface_index, bridge_name))

        ip.close()

    @staticmethod
    def remove_interface(interface_name: str) -> None:
        """
        Remove an interface from the host.

        Args:
            interface_name (str): The name of the interface to remove.

        Returns:
            None
        """
        from pyroute2 import IPRoute
        ip = IPRoute()

        logging.debug("Searching for interface `%s`..." % interface_name)

        # Search the interface
        link_indexes = ip.link_lookup(ifname=interface_name)
        # If not present, raise an error
        if not link_indexes:
            logging.debug("Interface `%s` not found, exiting..." % interface_name)
            return

        link_index = link_indexes[0]
        logging.debug("Interface found with ID = %d" % link_index)

        logging.debug("Removing interface with ID = %d" % link_index)
        ip.link(
            "del",
            index=link_index
        )

        ip.close()

    @staticmethod
    def get_iptables_version() -> str:
        """
        Return the iptables version.

        Returns:
            str: The iptables version.
        """
        return os.popen("/sbin/iptables --version").read().strip()



import logging
import re
from functools import partial
from multiprocessing.dummy import Pool
from typing import List, Union, Dict

import docker
import docker.models.networks
import progressbar
from docker import DockerClient
from docker import types

from ..docker.DockerPlugin import PLUGIN_NAME
from ... import utils
from ...exceptions import PrivilegeError
from ...model.ExternalLink import ExternalLink
from ...model.Lab import Lab
from ...model.Link import BRIDGE_LINK_NAME, Link
from ...os.Networking import Networking
from ...setting.Setting import Setting


class DockerLink(object):
    """The class responsible for deploying Kathara collision domains as Docker networks and interact with them."""
    __slots__ = ['client']

    def __init__(self, client: DockerClient) -> None:
        self.client: DockerClient = client

    def deploy_links(self, lab: Lab, selected_links: Dict[str, Link] = None) -> None:
        """Deploy all the lab collision domains as Docker networks.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_links (Dict[str, Link]): Keys are Link names, values are Link objects.

        Returns:
            None
        """
        links = selected_links.items() if selected_links else lab.links.items()

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            link_pool = Pool(pool_size)

            items = utils.chunk_list(links, pool_size)
            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deploying collision domains... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(links)
                )

            for chunk in items:
                link_pool.map(func=partial(self._deploy_link, progress_bar), iterable=chunk)

            if utils.CLI_ENV:
                progress_bar.finish()

        # Create a docker bridge link in the lab object and assign the Docker Network object associated to it.
        docker_bridge = self.get_docker_bridge()
        link = lab.get_or_new_link(BRIDGE_LINK_NAME)
        link.api_object = docker_bridge

    def _deploy_link(self, progress_bar: progressbar.ProgressBar, link_item: (str, Link)) -> None:
        """Deploy the collision domain contained in the link_item as a Docker network.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            link_item (Tuple[str, Link]): A tuple composed by the name of the collision domain and a Link object

        Returns:
            None
        """
        (_, link) = link_item

        if link.name == BRIDGE_LINK_NAME:
            return

        self.create(link)

        if progress_bar is not None:
            progress_bar += 1

    def create(self, link: Link) -> None:
        """Create a Docker network representing the collision domain object and assign it to link.api_object.

        It also connect external collision domains, if present.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.

        Returns:
            None
        """
        # Reserved name for bridged connections, ignore.
        if link.name == BRIDGE_LINK_NAME:
            return

        # If a network with the same name exists, return it instead of creating a new one.
        link_name = self.get_network_name(link.name)
        network_objects = self.get_links_api_objects_by_filters(link_name=link_name)
        if network_objects:
            link.api_object = network_objects.pop()
        else:
            network_ipam_config = docker.types.IPAMConfig(driver='null')

            user_label = "shared_cd" if Setting.get_instance().shared_cd else utils.get_current_user_name()
            link.api_object = self.client.networks.create(name=link_name,
                                                          driver=PLUGIN_NAME,
                                                          check_duplicate=True,
                                                          ipam=network_ipam_config,
                                                          labels={"lab_hash": link.lab.hash,
                                                                  "name": link.name,
                                                                  "user": user_label,
                                                                  "app": "kathara",
                                                                  "external": ";".join([x.get_full_name()
                                                                                        for x in link.external
                                                                                        ]
                                                                                       )
                                                                  }
                                                          )

        if link.external:
            logging.debug("External Interfaces required, connecting them...")
            self._attach_external_interfaces(link.external, link.api_object)

    def undeploy(self, lab_hash: str) -> None:
        """Undeploy all the collision domains of the scenario specified by lab_hash.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.

        Returns:
            None
        """
        links = self.get_links_api_objects_by_filters(lab_hash=lab_hash)
        for item in links:
            item.reload()
        links = [item for item in links if len(item.containers) <= 0]

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            links_pool = Pool(pool_size)

            items = utils.chunk_list(links, pool_size)

            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deleting collision domains... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(links)
                )

            for chunk in items:
                links_pool.map(func=partial(self._undeploy_link, progress_bar), iterable=chunk)

            if utils.CLI_ENV:
                progress_bar.finish()

    def wipe(self, user: str = None) -> None:
        """Undeploy all the Docker networks of the specified user. If user is None, it undeploy all the Docker networks.

        Args:
            user (str): The name of a current user on the host

        Returns:
            None
        """
        user_label = "shared_cd" if Setting.get_instance().shared_cd else user
        links = self.get_links_api_objects_by_filters(user=user_label)
        for item in links:
            item.reload()
        links = [item for item in links if len(item.containers) <= 0]

        pool_size = utils.get_pool_size()
        links_pool = Pool(pool_size)

        items = utils.chunk_list(links, pool_size)

        for chunk in items:
            links_pool.map(func=partial(self._undeploy_link, None), iterable=chunk)

    def _undeploy_link(self, progress_bar: progressbar.ProgressBar, network: docker.models.networks.Network) -> None:
        """Undeploy a Docker network.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            network (docker.models.networks.Network): The Docker network to undeploy.

        Returns:
            None
        """
        self._delete_link(network)

        if progress_bar is not None:
            progress_bar += 1

    def get_docker_bridge(self) -> Union[None, docker.models.networks.Network]:
        """Return the Docker bridged network.

        Returns:
            Union[None, docker.models.networks.Network]: The Docker bridged network if exists, else None
        """
        bridge_list = self.client.networks.list(names="bridge")
        return bridge_list.pop() if bridge_list else None

    def get_links_api_objects_by_filters(self, lab_hash: str = None, link_name: str = None, user: str = None) -> \
            List[docker.models.networks.Network]:
        """Return the Docker networks specified by lab_hash, machine_name and user.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the networks in the scenario.
            link_name (str): The name of a network. If specified, return the specified networks of the scenario.
            user (str): The name of a user on the host. If specified, return only the networks of the user.

        Returns:
            List[docker.models.networks.Network]: A list of Docker networks.
        """
        filters = {"label": ["app=kathara"]}
        if user:
            filters["label"].append("user=%s" % user)
        if lab_hash:
            filters["label"].append("lab_hash=%s" % lab_hash)
        if link_name:
            filters["name"] = link_name

        return self.client.networks.list(filters=filters)

    def _attach_external_interfaces(self, external_links: List[ExternalLink],
                                    network: docker.models.networks.Network) -> None:
        """Attach an external collision domain to a Docker network.

        Args:
            external_links (Kathara.model.ExternalLink): A Kathara external collision domain. It is used to create
            a collision domain attached to a host interface.
            network (docker.models.networks.Network): A Docker network.

        Returns:
            None
        """
        for external_link in external_links:
            (name, vlan) = external_link.get_name_and_vlan()

            interface_index = Networking.get_or_new_interface(external_link.interface, name, vlan)
            Networking.attach_interface_to_bridge(interface_index, self._get_bridge_name(network))

    @staticmethod
    def _get_bridge_name(network: docker.models.networks.Network) -> str:
        """Return the name of the host bridge associated to the Docker network.

        Args:
            network (docker.models.networks.Network): A Docker network.

        Returns:
            str: The name of the Docker bridge in the format "kt-<network.id[:12]>".
        """
        return "kt-%s" % network.id[:12]

    @staticmethod
    def get_network_name(name: str) -> str:
        """Return the name of a Docker network.

        Args:
            name (str): The name of a Kathara collision domain.

        Returns:
            str: The name of the Docker network in the format "|net_prefix|_|username_prefix|_|name|".
                If shared collision domains, the format is: "|net_prefix|_|lab_hash|".
        """
        username_prefix = "_%s" % utils.get_current_user_name() if not Setting.get_instance().shared_cd else ""
        return "%s%s_%s" % (Setting.get_instance().net_prefix, username_prefix, name)

    @staticmethod
    def _delete_link(network: docker.models.networks.Network) -> None:
        """Delete a Docker network.

        Args:
            network (docker.models.networks.Network): A Docker network.
        """
        external_label = network.attrs['Labels']["external"]
        external_links = external_label.split(";") if external_label else None

        if external_links:
            for external_link in external_links:
                if re.search(r"^\w+\.\d+$", external_link):
                    if utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2):
                        if utils.is_admin():
                            # Only remove VLAN interfaces, physical ones cannot be removed.
                            Networking.remove_interface(external_link)
                        else:
                            raise PrivilegeError("You must be root in order to delete a VLAN Interface.")
                    else:
                        raise OSError("VLAN Interfaces are only available on UNIX systems.")

        network.remove()



import json
import signal
from typing import Callable

import pyuv
from kubernetes.stream.ws_client import RESIZE_CHANNEL

from ....foundation.manager.terminal.Terminal import Terminal
from ....foundation.manager.terminal.terminal_utils import get_terminal_size_windows
from ....utils import exec_by_platform


class KubernetesWSTerminal(Terminal):
    def _start_external(self) -> None:
        self._external_terminal = pyuv.Timer(self._loop)
        self._external_terminal.start(self._read_external_terminal(), 0, 0.001)

    def _on_close(self) -> None:
        def unix_close():
            self._system_stdin.set_mode(0)

            self._resize_signal.close()

        exec_by_platform(unix_close, lambda: None, unix_close)

    def _write_on_external_terminal(self) -> Callable:
        def write_on_external_terminal(handle, data, error):
            self.handler.write_stdin(data)

        return write_on_external_terminal

    def _read_external_terminal(self) -> Callable:
        def read_external_terminal(timer_handle):
            if not self.handler.is_open() and not self._external_terminal.closed:
                self.close()
                return

            data = None
            if self.handler.peek_stdout():
                data = self.handler.read_stdout()
            elif self.handler.peek_stderr():
                data = self.handler.read_stderr()

            if data:
                self._system_stdout.write(data.encode('utf-8'))

                if data.strip() == '\r\nexit\r\n':
                    self.close()

        return read_external_terminal

    def _resize_terminal(self) -> None:
        def resize_unix():
            def resize_terminal(signal_handle, signal_num):
                w, h = self._system_stdin.get_winsize()
                self.handler.write_channel(RESIZE_CHANNEL, json.dumps({"Height": h, "Width": w}))

            self._resize_signal = pyuv.Signal(self._loop)
            self._resize_signal.start(resize_terminal, signal.SIGWINCH)

            # Run first time to set the proper terminal size
            resize_terminal(None, None)

        def resize_windows():
            w, h = get_terminal_size_windows()
            self.handler.write_channel(RESIZE_CHANNEL, json.dumps({"Height": h, "Width": w}))

        exec_by_platform(resize_unix, resize_windows, resize_unix)



import argparse
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Link import BRIDGE_LINK_NAME
from ...parser.netkit.LabParser import LabParser
from ...strings import strings, wiki_description
from ...trdparty.curses.curses import Curses


class LinfoCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara linfo',
            description=strings['linfo'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the lab.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-l', '--live',
            required=False,
            action='store_true',
            help='Live mode, can be used only when a lab is launched.'
        )

        group.add_argument(
            '-c', '--conf',
            required=False,
            action='store_true',
            help='Read information from lab.conf.'
        )

        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=False,
            help='Show only information about a specified device.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)
        lab_hash = utils.generate_urlsafe_hash(lab_path)

        if args['live']:
            if args['name']:
                self._get_machine_live_info(lab_hash, args['name'])
            else:
                self._get_lab_live_info(lab_hash)

            return

        if args['conf']:
            if args['name']:
                print(Kathara.get_instance().get_formatted_machine_info(args['name'], lab_hash))
            else:
                self._get_conf_info(lab_path)

            return

        if args['name']:
            print(Kathara.get_instance().get_formatted_machine_info(args['name'], lab_hash))
        else:
            lab_info = Kathara.get_instance().get_formatted_lab_info(lab_hash)

            print(next(lab_info))

    @staticmethod
    def _get_machine_live_info(lab_hash, machine_name):
        Curses.get_instance().init_window()

        try:
            while True:
                Curses.get_instance().print_string(
                    Kathara.get_instance().get_formatted_machine_info(machine_name, lab_hash)
                )
        finally:
            Curses.get_instance().close()

    @staticmethod
    def _get_lab_live_info(lab_hash: str) -> None:
        lab_info = Kathara.get_instance().get_formatted_lab_info(lab_hash)

        Curses.get_instance().init_window()

        try:
            while True:
                Curses.get_instance().print_string(next(lab_info))
        except StopIteration:
            pass
        finally:
            Curses.get_instance().close()

    @staticmethod
    def _get_conf_info(lab_path: str) -> None:
        print(utils.format_headers("Lab Information"))

        lab = LabParser.parse(lab_path)
        lab_meta_information = str(lab)

        if lab_meta_information:
            print(lab_meta_information)
            print(utils.format_headers())

        n_machines = len(lab.machines)
        n_links = len(lab.links) if BRIDGE_LINK_NAME not in lab.links else len(lab.links) - 1

        print("There are %d devices." % n_machines)
        print("There are %d collision domains." % n_links)

        print(utils.format_headers())



from typing import Optional, Iterable

from kubernetes import client
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException

from ...model.Lab import Lab


class KubernetesNamespace(object):
    """Class responsible for interacting with Kubernetes namespaces."""

    __slots__ = ['client']

    def __init__(self) -> None:
        self.client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

    def create(self, lab: Lab) -> Optional[client.V1Namespace]:
        """Return a Kubernetes namespace from a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.

        Returns:
            Optional[client.V1Namespace]: The Kubernetes namespace for the network scenario.
        """
        namespace_definition = client.V1Namespace(metadata=client.V1ObjectMeta(name=lab.hash,
                                                                               labels={'app': 'kathara'}
                                                                               )
                                                  )

        try:
            self.client.create_namespace(namespace_definition)
        except ApiException:
            return None

    def undeploy(self, lab_hash: str = None) -> None:
        """Delete the Kubernetes namespace corresponding to the lab_hash.

        Args:
            lab_hash (str): The hash of a Kathara network scenario.

        Returns:
            None
        """
        try:
            self.client.delete_namespace(lab_hash)
        except ApiException:
            return

    def wipe(self) -> None:
        """Delete all the Kathara Kubernetes namespaces.

        Returns:
            None
        """
        namespaces = self.get_all()

        for namespace in namespaces:
            self.client.delete_namespace(namespace.metadata.name)

    def get_all(self) -> Iterable[client.V1Namespace]:
        """Return an Iterable containing all the Kubernetes namespaces relatively to Kathara.

        Returns:
            Iterable[client.V1Namespace]: an Iterable containing all the Kubernetes namespaces relatively to Kathara.
        """
        return self.client.list_namespace(label_selector="app=kathara").items



from __future__ import annotations

import json
import logging
import os
import time
from typing import Any, Dict, Optional

from .. import utils
from .. import version
from ..connectors.GitHubApi import GitHubApi
from ..exceptions import HTTPConnectionError, SettingsError
from ..foundation.setting.SettingsAddon import SettingsAddon
from ..foundation.setting.SettingsAddonFactory import SettingsAddonFactory

AVAILABLE_DEBUG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "EXCEPTION"]
AVAILABLE_MANAGERS = ["docker", "kubernetes"]

ONE_WEEK = 604800

DEFAULTS = {
    "image": 'kathara/quagga',
    "manager_type": 'docker',
    "terminal": utils.exec_by_platform(lambda: '/usr/bin/xterm', lambda: '', lambda: 'Terminal'),
    "open_terminals": True,
    "device_shell": '/bin/bash',
    "net_prefix": 'kathara',
    "device_prefix": 'kathara',
    "debug_level": 'INFO',
    "print_startup_log": True,
    "enable_ipv6": False
}


class Setting(object):
    """Class responsible for interacting with Kathara Settings."""

    __slots__ = ['image', 'manager_type', 'terminal', 'open_terminals', 'device_shell', 'net_prefix',
                 'device_prefix', 'debug_level', 'print_startup_log', 'enable_ipv6', 'last_checked', 'addons']

    SETTING_FOLDER: str = None
    SETTING_PATH: str = None

    __instance: Setting = None

    @staticmethod
    def get_instance() -> Setting:
        """Return an instance of Setting.

        Returns:
            Kathara.setting.Setting: An instanche of Setting.
        """
        if Setting.__instance is None:
            Setting()

        return Setting.__instance

    def __init__(self) -> None:
        if Setting.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            Setting.SETTING_FOLDER = os.path.join(utils.get_current_user_home(), ".config")
            Setting.SETTING_PATH = os.path.join(Setting.SETTING_FOLDER, "kathara.conf")

            # Load default settings to use
            for (name, value) in DEFAULTS.items():
                setattr(self, name, value)

            self.addons: Optional[SettingsAddon] = None
            self.last_checked: float = time.time() - ONE_WEEK

            self.load()

            Setting.__instance = self

    def __getattr__(self, item: str) -> Any:
        return self.addons.get(item)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__slots__:
            super(Setting, self).__setattr__(name, value)
            return

        setattr(self.addons, name, value)

    def load(self) -> None:
        """Load the Kathara settings from disk.

        Returns:
            None
        """
        if not os.path.exists(Setting.SETTING_PATH):  # If settings file don't exist, create with defaults
            if not os.path.isdir(Setting.SETTING_FOLDER):  # Create .config folder if doesn't exists, create it
                os.mkdir(Setting.SETTING_FOLDER)

            self.load_settings_addon()  # Load default manager addons
            self.save()

            def unix_permissions():
                (uid, gid) = utils.get_current_user_uid_gid()

                os.chmod(Setting.SETTING_PATH, 0o600)
                os.chown(Setting.SETTING_PATH, uid, gid)

            # If Linux or Mac, set the right permissions and ownership to the settings file.
            utils.exec_by_platform(unix_permissions, lambda: None, unix_permissions)
        else:  # If file exists, read it and check values
            settings = {}
            with open(Setting.SETTING_PATH, 'r') as settings_file:
                try:
                    settings = json.load(settings_file)
                except ValueError:
                    raise SettingsError("Not a valid JSON.")

            for name, value in settings.items():
                if hasattr(self, name):
                    setattr(self, name, value)

            self.load_settings_addon()  # Manager may be changed with loaded settings, reload addon
            self.addons.load(settings)  # Load values into the addon object

    @staticmethod
    def wipe() -> None:
        """Remove the all saved settings from disk.

        Returns:
            None
        """
        if os.path.exists(Setting.SETTING_PATH):
            os.remove(Setting.SETTING_PATH)

    def save(self) -> None:
        """Saves settings to a config.json file in the Kathara path.

        Returns:
            None
        """
        to_save = self.addons.merge(self._to_dict())

        with open(Setting.SETTING_PATH, 'w') as settings_file:
            settings_file.write(json.dumps(to_save, indent=True))

    def check(self) -> None:
        """Chek if Kathara is correctly working.

        Check if the the selected manager and terminal are available. Check the presence of Kathara updates.
        Check the correctness and validity of the net_prefix, device_prefix and debug level.

        Returns:
            None
        """
        self.check_manager()

        self.check_terminal()

        current_time = time.time()
        # After 1 week, check if a new Kathara version has been released.
        if current_time - self.last_checked > ONE_WEEK:
            logging.debug(utils.format_headers("Checking Updates"))
            checked = True

            try:
                logging.debug("Checking Kathara release...")

                latest_remote_release = GitHubApi.get_release_information()
                latest_version = latest_remote_release["tag_name"]

                if version.less_than(version.CURRENT_VERSION, latest_version):
                    logging.info("A new version of Kathara has been released.")
                    logging.info("Current: %s - Latest: %s" % (version.CURRENT_VERSION, latest_version))
                    logging.info("Please update it from https://github.com/KatharaFramework/Kathara")
            except HTTPConnectionError:
                logging.debug("Connection to GitHub failed, passing...")
                checked = False

            if checked:
                self.last_checked = current_time
                self.save()

            logging.debug(utils.format_headers())

        try:
            utils.re_search_fail(r"^[a-z]+_?[a-z_]+$", self.net_prefix)
        except ValueError:
            raise SettingsError("Networks Prefix must only contain lowercase letters and underscore.")

        try:
            utils.re_search_fail(r"^[a-z]+_?[a-z_]+$", self.device_prefix)
        except ValueError:
            raise SettingsError("Device Prefix must only contain lowercase letters and underscore.")

        if self.debug_level not in AVAILABLE_DEBUG_LEVELS:
            raise SettingsError("Debug Level must be one of the following: %s." % (", ".join(AVAILABLE_DEBUG_LEVELS)))

    def check_manager(self) -> None:
        """Check if the selected manager is available.

        Returns:
            None

        Raises:
            SettingsError: "Manager Type not allowed."
        """
        from ..manager.Kathara import Kathara
        managers = Kathara.get_available_managers_name()

        if self.manager_type not in managers.keys():
            raise SettingsError("Manager Type not allowed.")

    def check_image(self, image: str = None) -> None:
        """Check if the specified image is valid.

        Args:
            image (str): The name of the image to check. If None, check the default image.

        Returns:
            None

        Raises:
            ConnectionError: The image is not locally available and there is no connection to a remote image repository.
            Exception: The image is not found.
        """
        image = self.image if not image else image

        # Required to import here because otherwise there is a cyclic dependency
        from ..manager.Kathara import Kathara
        Kathara.get_instance().check_image(image)

    def check_terminal(self, terminal: str = None) -> bool:
        """Check that the selected terminal is available.

        Args:
            terminal (str): The selected terminal path. If None, check the availability of the default terminal.

        Returns:
            bool: True if the selected terminal is TMUX (that do not require path), else False.

        """
        terminal = self.terminal if not terminal else terminal

        # Skip check for TMUX (special value)
        if terminal == "TMUX":
            return True

        def check_unix():
            return os.path.isfile(terminal) and os.access(terminal, os.X_OK)

        def check_osx():
            import appscript
            import aem
            try:
                appscript.app(terminal)
            except aem.findapp.ApplicationNotFoundError:
                return False

            return True

        if not utils.exec_by_platform(check_unix, lambda: True, check_osx):
            raise SettingsError("Terminal Emulator `%s` not valid! Install it before using it." % terminal)

        return True

    def load_settings_addon(self) -> None:
        """Load a setting addon to the base settings.

        Returns:
            None
        """
        self.addons = SettingsAddonFactory().create_instance(class_args=(self.manager_type.capitalize(),))

    def _to_dict(self) -> Dict[str, Any]:
        return {"image": self.image,
                "manager_type": self.manager_type,
                "terminal": self.terminal,
                "open_terminals": self.open_terminals,
                "device_shell": self.device_shell,
                "net_prefix": self.net_prefix,
                "device_prefix": self.device_prefix,
                "debug_level": self.debug_level,
                "print_startup_log": self.print_startup_log,
                "enable_ipv6": self.enable_ipv6,
                "last_checked": self.last_checked
                }



#!/usr/bin/env python3

import argparse
import logging
import multiprocessing
import os
import sys

import coloredlogs

from Kathara import utils
from Kathara.auth.PrivilegeHandler import PrivilegeHandler
from Kathara.exceptions import SettingsError, DockerDaemonConnectionError, ClassNotFoundError
from Kathara.foundation.cli.command.CommandFactory import CommandFactory
from Kathara.setting.Setting import Setting
from Kathara.strings import formatted_strings
from Kathara.version import CURRENT_VERSION

description_msg = """kathara [-h] [-v] <command> [<args>]

Possible Kathara commands are:\n
%s
""" % formatted_strings()


class KatharaEntryPoint(object):
    def __init__(self) -> None:
        parser = argparse.ArgumentParser(
            description='A network emulation tool.',
            usage=description_msg,
            add_help=False
        )

        parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        parser.add_argument(
            'command',
            help='Command to run.',
            nargs="?"
        )

        parser.add_argument(
            "-v", "--version",
            action="store_true",
            help='Print the current Kathara version.',
            required=False
        )

        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])

        if args.version:
            print('Current version: %s' % CURRENT_VERSION)
            sys.exit(0)

        if args.command is None or not args.command.islower():
            parser.print_help()
            sys.exit(1)

        try:
            # Check settings
            Setting.get_instance().check()
        except (SettingsError, DockerDaemonConnectionError) as e:
            logging.critical(str(e))
            sys.exit(1)

        try:
            command_object = CommandFactory().create_instance(class_args=(args.command.capitalize(),))
        except ClassNotFoundError:
            logging.error('Unrecognized command `%s`.\n' % args.command)
            parser.print_help()
            sys.exit(1)
        except ImportError as e:
            logging.critical("`%s` is not installed in your system\n" % e.name)
            sys.exit(1)

        try:
            current_path = os.getcwd()
            command_object.run(current_path, sys.argv[2:])
        except KeyboardInterrupt:
            if args.command not in ['exec', 'linfo', 'list', 'settings']:
                logging.warning("You interrupted Kathara during a command. The system may be in an inconsistent "
                                "state!\n")
                logging.warning("If you encounter any problem please run `kathara wipe`.")
            sys.exit(0)
        except Exception as e:
            if Setting.get_instance().debug_level == "EXCEPTION":
                logging.exception(str(e) + '\n')
            else:
                logging.critical(str(e) + '\n')
            sys.exit(1)


if __name__ == '__main__':
    multiprocessing.freeze_support()

    utils.CLI_ENV = True
    utils.check_python_version()

    utils.exec_by_platform(PrivilegeHandler.get_instance().drop_privileges, lambda: None, lambda: None)

    try:
        debug_level = Setting.get_instance().debug_level
        debug_level = debug_level if debug_level != "EXCEPTION" else "DEBUG"
    except SettingsError:
        debug_level = "DEBUG"

    coloredlogs.install(fmt='%(levelname)s - %(message)s', level=debug_level)

    KatharaEntryPoint()



import argparse
from typing import List

from .LcleanCommand import LcleanCommand
from .LstartCommand import LstartCommand
from ... import utils
from ...foundation.cli.command.Command import Command
from ...strings import strings, wiki_description


class LrestartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lrestart',
            description=strings['lrestart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=True,
            help='Start the lab without opening terminal windows.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the lab opening terminal windows.'
        )
        group.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the devices in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the lab.'
        )
        self.parser.add_argument(
            '-F', '--force-lab',
            dest='force_lab',
            required=False,
            action='store_true',
            help='Force the lab to start without a lab.conf or lab.dep file.'
        )
        self.parser.add_argument(
            '-l', '--list',
            required=False,
            action='store_true',
            help='Show information about running devices after the lab has been started.'
        )
        self.parser.add_argument(
            '-o', '--pass',
            dest='options',
            metavar="OPTION",
            nargs='*',
            required=False,
            help="Apply options to all devices of a lab during startup."
        )
        self.parser.add_argument(
            '--xterm',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        self.parser.add_argument(
            '-H', '--no-hosthome',
            dest="no_hosthome",
            action="store_const",
            const=False,
            help='/hosthome dir will not be mounted inside the devices.'
        )
        self.parser.add_argument(
            '-S', '--no-shared',
            dest="no_shared",
            action="store_const",
            const=False,
            help='/shared dir will not be mounted inside the devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        lclean_argv = ['-d', args['directory']] if args['directory'] else []

        LcleanCommand().run(lab_path, lclean_argv)
        LstartCommand().run(lab_path, argv)



import hashlib
import json
import logging
import re
import shlex
import sys
import uuid
from functools import partial
from multiprocessing.dummy import Pool
from typing import Optional, Set, List, Union, Dict, Generator, Tuple, Any

import progressbar
from kubernetes import client
from kubernetes.client.api import apps_v1_api
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream

from .KubernetesConfigMap import KubernetesConfigMap
from .KubernetesNamespace import KubernetesNamespace
from ... import utils
from ...exceptions import MachineAlreadyExistsError
from ...model.Lab import Lab
from ...model.Machine import Machine
from ...setting.Setting import Setting

RP_FILTER_NAMESPACE = "net.ipv4.conf.%s.rp_filter"

# Known commands that each container should execute
# Run order: shared.startup, machine.startup and machine.startup_commands
STARTUP_COMMANDS = [
    # If execution flag file is found, abort (this means that postStart has been called again)
    # If not flag the startup execution with a file
    "if [ -f \"/tmp/post_start\" ]; then exit; else touch /tmp/post_start; fi",

    "{sysctl_commands}",

    # Removes /etc/bind already existing configuration from k8s internal DNS
    "rm -Rf /etc/bind/*",

    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Docker inside the container.
    # In this way, they can be overwritten by custom user files.
    "umount /etc/resolv.conf",
    "umount /etc/hosts",

    # Parse hostlab.b64
    "base64 -d /tmp/kathara/hostlab.b64 > /hostlab.tar.gz",
    # Extract hostlab.tar.gz data into /
    "tar xmfz /hostlab.tar.gz -C /; rm -f hostlab.tar.gz",

    # Copy the machine folder (if present) from the hostlab directory into the root folder of the container
    # In this way, files are all replaced in the container root folder
    "if [ -d \"/hostlab/{machine_name}\" ]; then "
    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf -); fi",

    # Give proper permissions to /var/www
    "if [ -d \"/var/www\" ]; then "
    "chmod -R 777 /var/www/*; fi",

    # Give proper permissions to Quagga files (if present)
    "if [ -d \"/etc/quagga\" ]; then "
    "chown quagga:quagga /etc/quagga/*",
    "chmod 640 /etc/quagga/*; fi",

    # Give proper permissions to FRR files (if present)
    "if [ -d \"/etc/frr\" ]; then "
    "chown frr:frr /etc/frr/*",
    "chmod 640 /etc/frr/*; fi",

    # If shared.startup file is present
    "if [ -f \"/hostlab/shared.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/shared.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/shared.startup",
    "/hostlab/shared.startup &> /var/log/shared.log; fi",

    # If .startup file is present
    "if [ -f \"/hostlab/{machine_name}.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/{machine_name}.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/{machine_name}.startup",
    "/hostlab/{machine_name}.startup &> /var/log/startup.log; fi",

    # Remove the Kubernetes' default gateway which points to the eth0 interface and causes problems sometimes.
    "route del default dev eth0 || true",

    # Placeholder for user commands
    "{machine_commands}"
]

SHUTDOWN_COMMANDS = [
    # If machine.shutdown file is present
    "if [ -f \"/hostlab/{machine_name}.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/{machine_name}.shutdown; /hostlab/{machine_name}.shutdown; fi",

    # If shared.shutdown file is present
    "if [ -f \"/hostlab/shared.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/shared.shutdown; /hostlab/shared.shutdown; fi"
]


class KubernetesMachine(object):
    """Class responsible for managing Kathara devices representation in Kubernetes."""
    __slots__ = ['client', 'core_client', 'kubernetes_config_map', 'kubernetes_namespace']

    def __init__(self, kubernetes_namespace: KubernetesNamespace) -> None:
        self.client: apps_v1_api.AppsV1Api = apps_v1_api.AppsV1Api()
        self.core_client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

        self.kubernetes_config_map: KubernetesConfigMap = KubernetesConfigMap()

        self.kubernetes_namespace: KubernetesNamespace = kubernetes_namespace

    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """Deploy all the devices contained in lab.machines.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_machines (Set[str]): A set containing the name of the devices to deploy.

        Returns:
            None
        """

        machines = {k: v for (k, v) in lab.machines.items() if k in selected_machines}.items() if selected_machines \
            else lab.machines.items()

        privileged = lab.general_options['privileged_machines'] if 'privileged_machines' in lab.general_options \
            else False
        if privileged:
            logging.warning('Privileged option is not supported on Megalos. It will be ignored.')

        progress_bar = None
        if utils.CLI_ENV:
            progress_bar = progressbar.ProgressBar(
                widgets=['Deploying devices... ', progressbar.Bar(),
                         ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                redirect_stdout=True,
                max_value=len(machines)
            )

        # Deploy all lab machines.
        # If there is no lab.dep file, machines can be deployed using multithreading.
        # If not, they're started sequentially
        if not lab.has_dependencies:
            pool_size = utils.get_pool_size()
            machines_pool = Pool(pool_size)

            items = utils.chunk_list(machines, pool_size)

            for chunk in items:
                machines_pool.map(func=partial(self._deploy_machine, progress_bar),
                                  iterable=chunk
                                  )
        else:
            for item in machines:
                self._deploy_machine(progress_bar, item)

        if utils.CLI_ENV:
            progress_bar.finish()

    def _deploy_machine(self, progress_bar: progressbar.ProgressBar, machine_item: Tuple[str, Machine]) -> None:
        """Deploy a Kubernetes deployment from the Kathara device contained in machine_item.

        Args:
           progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
           machine_item (Tuple[str, Machine]): A tuple composed by the name of the device and a device object

        Returns:
           None
        """
        (_, machine) = machine_item

        self.create(machine)

        if progress_bar is not None:
            progress_bar += 1

    def create(self, machine: Machine) -> None:
        """Create a Kubernetes deployment and a PoD representing the device and assign it to machine.api_object.

        Args:
            machine (Kathara.model.Machine.Machine): a Kathara device.

        Returns:
            None
        """
        logging.debug("Creating device `%s`..." % machine.name)

        # Get the general options into a local variable (just to avoid accessing the lab object every time)
        options = machine.lab.general_options

        # If bridged is defined for the device, throw a warning.
        if "bridged" in options or machine.meta['bridged']:
            logging.warning('Bridged option is not supported on Megalos. It will be ignored.')

        # If any exec command is passed in command line, add it.
        if "exec" in options:
            machine.add_meta("exec", options["exec"])

        # Sysctl params to pass to the container creation
        sysctl_parameters = {RP_FILTER_NAMESPACE % x: 0 for x in ["all", "default", "lo"]}

        sysctl_parameters["net.ipv4.ip_forward"] = 1
        sysctl_parameters["net.ipv4.icmp_ratelimit"] = 0

        if machine.is_ipv6_enabled():
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 1
            sysctl_parameters["net.ipv6.icmp.ratelimit"] = 0
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 0
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 0

        # Merge machine sysctls
        machine.meta['sysctls'] = {**sysctl_parameters, **machine.meta['sysctls']}

        machine.add_meta('real_name', self.get_deployment_name(machine.name))

        try:
            config_map = self.kubernetes_config_map.deploy_for_machine(machine)
            machine_definition = self._build_definition(machine, config_map)

            machine.api_object = self.client.create_namespaced_deployment(body=machine_definition,
                                                                          namespace=machine.lab.hash
                                                                          )
        except ApiException as e:
            if e.status == 409 and 'Conflict' in e.reason:
                raise MachineAlreadyExistsError("Device with name `%s` already exists." % machine.name)
            else:
                raise e

    @staticmethod
    def _build_definition(machine: Machine, config_map: client.V1ConfigMap) -> client.V1Deployment:
        """Return a Kubernetes deployment from a Kathara device and a Kubernetes ConfigMap.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.
            config_map (client.V1ConfigMap): A Kubernetes ConfigMap containing the tar data to upload on the deployment.

        Returns:
            client.V1Deployment: A Kubernetes deployment.
        """
        volume_mounts = []
        if config_map:
            # Define volume mounts for hostlab if a ConfigMap is defined.
            volume_mounts.append(client.V1VolumeMount(name="hostlab", mount_path="/tmp/kathara"))

        if Setting.get_instance().host_shared:
            volume_mounts.append(client.V1VolumeMount(name="shared", mount_path="/shared"))

        # Machine must be executed in privileged mode to run sysctls.
        security_context = client.V1SecurityContext(privileged=True)

        ports_info = machine.get_ports()
        container_ports = None
        if ports_info:
            container_ports = []
            for (host_port, protocol), guest_port in ports_info.items():
                container_ports.append(
                    client.V1ContainerPort(
                        name=str(uuid.uuid4()).replace('-', '')[0:15],
                        container_port=guest_port,
                        host_port=host_port,
                        protocol=protocol.upper()
                    )
                )

        resources = None
        memory = machine.get_mem()
        cpus = machine.get_cpu(multiplier=1000)
        if memory or cpus:
            limits = dict()
            if memory:
                limits["memory"] = memory.upper()
            if cpus:
                limits["cpu"] = "%dm" % cpus

            resources = client.V1ResourceRequirements(limits=limits)

        # postStart lifecycle hook is launched asynchronously by k8s master when the main container is Ready
        # On Ready state, the pod has volumes and network interfaces up, so this hook is used
        # to execute custom commands coming from .startup file and "exec" option
        # Build the final startup commands string
        sysctl_commands = "; ".join(["sysctl -w -q %s=%d" % item for item in machine.meta["sysctls"].items()])
        startup_commands_string = "; ".join(STARTUP_COMMANDS) \
            .format(machine_name=machine.name,
                    sysctl_commands=sysctl_commands,
                    machine_commands="; ".join(machine.startup_commands)
                    )

        post_start = client.V1Handler(
            _exec=client.V1ExecAction(
                command=[Setting.get_instance().device_shell, "-c", startup_commands_string]
            )
        )
        lifecycle = client.V1Lifecycle(post_start=post_start)

        env = [client.V1EnvVar("_MEGALOS_SHELL",
                               machine.meta["shell"] if "shell" in machine.meta else Setting.get_instance().device_shell
                               )
               ]

        container_definition = client.V1Container(
            name=machine.meta['real_name'],
            image=machine.get_image(),
            lifecycle=lifecycle,
            stdin=True,
            tty=True,
            image_pull_policy=Setting.get_instance().image_pull_policy,
            ports=container_ports,
            resources=resources,
            volume_mounts=volume_mounts,
            security_context=security_context,
            env=env
        )

        pod_annotations = {}
        network_interfaces = []
        for (idx, machine_link) in machine.interfaces.items():
            network_interfaces.append({
                "name": machine_link.api_object["metadata"]["name"],
                "namespace": machine.lab.hash,
                "interface": "net%d" % idx
            })
        pod_annotations["k8s.v1.cni.cncf.io/networks"] = json.dumps(network_interfaces)

        # Create labels (so Deployment can match them)
        pod_labels = {"name": machine.name,
                      "app": "kathara"
                      }

        pod_metadata = client.V1ObjectMeta(deletion_grace_period_seconds=0,
                                           annotations=pod_annotations,
                                           labels=pod_labels
                                           )

        # Add fake DNS just to override k8s one
        dns_config = client.V1PodDNSConfig(nameservers=["127.0.0.1"])

        volumes = []
        if config_map:
            # Hostlab is the lab base64 encoded .tar.gz of the machine files, deployed as a ConfigMap in the cluster
            # The base64 file is mounted into /tmp and it's extracted by the postStart hook
            volumes.append(client.V1Volume(
                name="hostlab",
                config_map=client.V1ConfigMapVolumeSource(
                    name=config_map.metadata.name
                )
            ))

        # Container /shared mounts in /home/shared folder
        if Setting.get_instance().host_shared:
            volumes.append(client.V1Volume(
                name="shared",
                host_path=client.V1HostPathVolumeSource(
                    path='/home/shared',
                    type='DirectoryOrCreate'
                )
            ))

        pod_spec = client.V1PodSpec(containers=[container_definition],
                                    hostname=machine.meta['real_name'],
                                    dns_policy="None",
                                    dns_config=dns_config,
                                    volumes=volumes
                                    )

        pod_template = client.V1PodTemplateSpec(metadata=pod_metadata, spec=pod_spec)
        selector_rules = client.V1LabelSelector(match_labels=pod_labels)
        deployment_spec = client.V1DeploymentSpec(replicas=1,
                                                  template=pod_template,
                                                  selector=selector_rules
                                                  )
        deployment_metadata = client.V1ObjectMeta(name=machine.meta['real_name'], labels=pod_labels)

        return client.V1Deployment(api_version="apps/v1",
                                   kind="Deployment",
                                   metadata=deployment_metadata,
                                   spec=deployment_spec
                                   )

    def undeploy(self, lab_hash: str, selected_machines: Optional[Set[str]] = None) -> None:
        """Undeploy all the running Kubernetes deployments and PoDs contained in the scenario defined by the lab_hash.

        If selected_machines is not None, undeploy only the specified devices.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        machines = self.get_machines_api_objects_by_filters(lab_hash=lab_hash)
        if selected_machines is not None and len(selected_machines) > 0:
            machines = [item for item in machines if item.metadata.labels["name"] in selected_machines]

        if len(machines) > 0:
            pool_size = utils.get_pool_size()
            machines_pool = Pool(pool_size)

            items = utils.chunk_list(machines, pool_size)

            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deleting devices... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(machines)
                )

            for chunk in items:
                machines_pool.map(func=partial(self._undeploy_machine, progress_bar),
                                  iterable=chunk
                                  )
            if utils.CLI_ENV:
                progress_bar.finish()

    def wipe(self) -> None:
        """Undeploy all the running Kubernetes deployments and PoDs.

        Returns:
            None
        """
        machines = self.get_machines_api_objects_by_filters()

        pool_size = utils.get_pool_size()
        machines_pool = Pool(pool_size)

        items = utils.chunk_list(machines, pool_size)

        for chunk in items:
            machines_pool.map(func=partial(self._undeploy_machine, None), iterable=chunk)

    def _undeploy_machine(self, progress_bar: progressbar.ProgressBar, pod_api_object: client.V1Pod) -> None:
        """Undeploy a Kubernetes pod.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            pod_api_object (client.V1Pod): The Kubernetes pod to undeploy.

        Returns:
            None
        """

        self._delete_machine(pod_api_object)

        if progress_bar is not None:
            progress_bar += 1

    def _delete_machine(self, pod_api_object: client.V1Pod) -> None:
        """Delete the Kubernetes deployment and PoD associated to pod_api_object.

        Args:
            pod_api_object (client.V1Pod): A Kubernetes PoD API object.

        Returns:
            None
        """
        machine_name = pod_api_object.metadata.labels["name"]
        machine_namespace = pod_api_object.metadata.namespace

        # Build the shutdown command string
        shutdown_commands_string = "; ".join(SHUTDOWN_COMMANDS).format(machine_name=machine_name)

        try:
            output = self.exec(machine_namespace,
                               machine_name,
                               command=[Setting.get_instance().device_shell, '-c', shutdown_commands_string],
                               )

            try:
                next(output)
            except StopIteration:
                pass

            deployment_name = self.get_deployment_name(machine_name)
            self.kubernetes_config_map.delete_for_machine(deployment_name, machine_namespace)

            self.client.delete_namespaced_deployment(name=deployment_name,
                                                     namespace=machine_namespace
                                                     )
        except ApiException:
            return

    def connect(self, lab_hash: str, machine_name: str, shell: Union[str, List[str]] = None, logs: bool = False) \
            -> None:
        """Open a stream to the Kubernetes PoD specified by machine_name using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device to connect.
            shell (Union[str, List[str]): The path to the desired shell.
            logs (bool): If True, print the logs of the startup command.

        Returns:
            None
        """
        pod = self.get_machine_api_object(lab_hash=lab_hash, machine_name=machine_name)

        if 'Running' not in pod.status.phase:
            raise Exception('Device `%s` is not ready.' % machine_name)

        if not shell:
            shell_env_value = self.get_env_var_value_from_pod(pod, "_MEGALOS_SHELL")
            shell = shlex.split(shell_env_value if shell_env_value else Setting.get_instance().device_shell)
        else:
            shell = shlex.split(shell) if type(shell) == str else shell

        logging.debug("Connect to device `%s` with shell: %s" % (machine_name, shell))

        if logs and Setting.get_instance().print_startup_log:
            exec_output = self.exec(lab_hash,
                                    machine_name,
                                    command="/bin/cat /var/log/shared.log /var/log/startup.log"
                                    )
            try:
                print("--- Startup Commands Log\n")
                while True:
                    (stdout, _) = next(exec_output)
                    stdout = stdout.decode('utf-8') if stdout else ""
                    sys.stdout.write(stdout)
            except StopIteration:
                print("\n--- End Startup Commands Log\n")
                pass

        resp = stream(self.core_client.connect_get_namespaced_pod_exec,
                      name=pod.metadata.name,
                      namespace=lab_hash,
                      command=shell,
                      stdout=True,
                      stderr=True,
                      stdin=True,
                      tty=True,
                      _preload_content=False
                      )

        from .terminal.KubernetesWSTerminal import KubernetesWSTerminal
        KubernetesWSTerminal(resp).start()

    @staticmethod
    def get_env_var_value_from_pod(pod: client.V1Pod, var_name: str) -> Optional[str]:
        """Return the value of an environment variable of the Kubernetes PoD.

        Args:
            pod (client.V1Pod): A Kubernetes PoD.
            var_name (str): The name of the environment variable.

        Returns:
            Optional[str]: The value of the environment variable.
        """
        containers = pod.spec.containers
        if not containers:
            return None

        # There is only one container definition in Megalos Pods
        container_definition = containers.pop()
        # Get env vars list
        env_vars = container_definition.env

        if not env_vars:
            return None

        # Iterate over the env vars and search the desired one
        for env_var in env_vars:
            if var_name == env_var.name:
                return env_var.value

        return None

    def exec(self, lab_hash: str, machine_name: str, command: Union[str, List], tty: bool = False, stdin: bool = False,
             stdin_buffer: List[Union[str, bytes]] = None, stderr: bool = False, is_stream: bool = False) \
            -> Generator[Tuple[bytes, bytes], None, None]:
        """Execute the command on the Kubernetes PoD specified by the lab_hash and the machine_name.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device.
            command (str): The command to execute.
            tty (bool): If True, open a new tty.
            stdin (bool): If True, open the stdin channel.
            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
            stderr (bool): If True, return the stderr.
            is_stream (bool): If True, return a generator with each line.
                If False, returns a generator with the complete output.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        command = shlex.split(command) if type(command) == str else command

        logging.debug("Executing command `%s` to device with name: %s" % (command, machine_name))

        try:
            # Retrieve the pod of current Deployment
            pod = self.get_machine_api_object(lab_hash=lab_hash, machine_name=machine_name)

            response = stream(self.core_client.connect_get_namespaced_pod_exec,
                              name=pod.metadata.name,
                              namespace=lab_hash,
                              command=command,
                              stdout=True,
                              stderr=stderr,
                              stdin=stdin,
                              tty=tty,
                              _preload_content=False
                              )
        except ApiException:
            return None

        if stdin_buffer is None:
            stdin_buffer = []

        result = {
            'stdout': '',
            'stderr': ''
        }

        while response.is_open():
            stdout = None
            stderr = None
            if response.peek_stdout():
                stdout = response.read_stdout()
                if not is_stream:
                    result['stdout'] += stdout
            if stderr and response.peek_stderr():
                stderr = response.read_stderr()
                if not is_stream:
                    result['stderr'] += stderr
            if stdin and stdin_buffer:
                param = stdin_buffer.pop(0)
                response.write_stdin(param)
                if len(stdin_buffer) <= 0:
                    break

            if is_stream and (stdout or stderr):
                yield stdout.encode('utf-8') if stdout else None, stderr.encode('utf-8') if stderr else None

        response.close()

        if not is_stream:
            yield result['stdout'].encode('utf-8'), result['stderr'].encode('utf-8')

    def copy_files(self, deployment: client.V1Deployment, path: str, tar_data: bytes) -> None:
        machine_name = deployment.metadata.labels["name"]
        machine_namespace = deployment.metadata.namespace

        exec_output = self.exec(machine_namespace,
                                machine_name,
                                command=['tar', 'xvfz', '-', '-C', path],
                                stdin=True,
                                stdin_buffer=[tar_data]
                                )

        try:
            next(exec_output)
        except StopIteration:
            pass

    def get_machines_api_objects_by_filters(self, lab_hash: str = None, machine_name: str = None) -> List[client.V1Pod]:
        """Return the List of Kubernetes PoD.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the Kubernetes PoD in the scenario.
            machine_name (str): The name of a device. If specified, return the specified Kubernetes PoD of the scenario.

        Returns:
            List[client.V1Pod]: A list of Kubernetes PoDs.
        """
        filters = ["app=kathara"]
        if machine_name:
            filters.append("name=%s" % machine_name)

        # Get all Kathara namespaces if lab_hash is None
        namespaces = list(map(lambda x: x.metadata.name, self.kubernetes_namespace.get_all())) \
            if not lab_hash else [lab_hash]

        machines = []
        for namespace in namespaces:
            machines.extend(self.core_client.list_namespaced_pod(namespace=namespace,
                                                                 label_selector=",".join(filters),
                                                                 timeout_seconds=9999
                                                                 ).items
                            )

        return machines

    def get_machine_api_object(self, lab_hash: str, machine_name: str) -> client.V1Pod:
        """Return the Kubernetes PoD specified by lab_hash and machine_name.

        Args:
            lab_hash (str): The hash of a network scenario.
            machine_name (str): The name of a device.

        Returns:
            client.V1Pod: A Kubernets PoD.
        """
        pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)

        logging.debug("Found pods: %s" % len(pods))

        if len(pods) != 1:
            raise Exception("Error getting the device `%s` inside the lab." % machine_name)
        else:
            return pods[0]

    def get_machine_info(self, machine_name: str, lab_hash: str = None) -> List[Dict[str, Any]]:
        """Return a list of dicts containing the devices info.

        Args:
            machine_name (str): The name of a device
            lab_hash (str): The hash of a network scenario. If specified, search only the devices in the corresponding
            scenario.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        machines_api_objects = self.get_machines_api_objects_by_filters(machine_name=machine_name, lab_hash=lab_hash)

        if not machines_api_objects:
            raise Exception("The specified device is not running.")

        all_stats = []
        for machine_api_object in machines_api_objects:
            all_stats.append(self._get_stats_by_machine(machine_api_object))

        return all_stats

    def get_machines_info(self, lab_hash: str, machine_name: str = None) -> Generator[Dict[str, Any], None, None]:
        """Return a generator containing the info of the specified device.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the devices in the scenario.
            machine_name (str): The name of a device. If specified, return the specified container of the scenario.

        Returns:
            Generator[Dict[str, Any], None, None]: A generator containing the info of the specified device.
        """
        while True:
            machines = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)

            if not machines:
                if not lab_hash:
                    raise Exception("No devices running.")
                else:
                    raise Exception("Lab is not started.")

            machines = sorted(machines, key=lambda x: x.metadata.labels["name"])

            machines_stats = []
            for machine in machines:
                machines_stats.append(self._get_stats_by_machine(machine))

            yield machines_stats

    def _get_stats_by_machine(self, pod_api_object: client.V1Pod) -> Dict[str, str]:
        """Return the stats of the specified Kubernetes PoD.

        Args:
            pod_api_object (client.V1Pod): A Docker container.

        Returns:
            Dict[str, Any]: A dict containing formatted Kathara device stats.
        """

        container_statuses = pod_api_object.status.container_statuses
        image_name = container_statuses[0].image.replace('docker.io/', '') if container_statuses else "N/A"

        return {
            "real_lab_hash": pod_api_object.metadata.namespace,
            "name": pod_api_object.metadata.labels["name"],
            "real_name": pod_api_object.metadata.name,
            "status": self._get_detailed_machine_status(pod_api_object),
            "image": image_name,
            "assigned_node": pod_api_object.spec.node_name
        }

    @staticmethod
    def _get_detailed_machine_status(pod_api_object: client.V1Pod) -> str:
        """Return a string containing the Kubernetes PoD status.

        Args:
            pod_api_object (client.V1Pod): A Kubernetes PoD.

        Returns:
            str: A string containing the Kubernetes PoD status.
        """
        container_statuses = pod_api_object.status.container_statuses

        if not container_statuses:
            return pod_api_object.status.phase

        container_status = container_statuses[0].state

        string_status = None
        if container_status.terminated is not None:
            string_status = container_status.terminated.reason if container_status.terminated.reason is not None \
                else "Terminating"
        elif container_status.waiting is not None:
            string_status = container_status.waiting.reason

        # In case the status contains an error message, split it to the first ": " and take the left part
        return string_status.split(': ')[0] if string_status is not None else pod_api_object.status.phase

    @staticmethod
    def get_deployment_name(name: str) -> str:
        """Return the name of the Kubernetes deployment corresponding to 'name'.

        Args:
            name (str): The name of a Kathara device.

        Returns:
            str: The name for the Kubernetes deployment.
        """
        suffix = ''
        # Underscore is replaced with -, but to keep name uniqueness append 8 chars of hash from the original name
        if '_' in name:
            suffix = '-%s' % hashlib.md5(name.encode('utf-8', errors='ignore')).hexdigest()[:8]
            name = name.replace('_', '-')

        machine_name = "%s-%s%s" % (Setting.get_instance().device_prefix, name, suffix)
        return re.sub(r'[^0-9a-z\-.]+', '', machine_name.lower())



import argparse
import logging
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description


class VcleanCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vclean',
            description=strings['vclean'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be cleaned.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_hash = utils.generate_urlsafe_hash("kathara_vlab")

        Kathara.get_instance().undeploy_lab(lab_hash,
                                            selected_machines={args['name']}
                                            )

        logging.info("Device `%s` deleted successfully!" % args['name'])



import logging

import requests

from ..exceptions import HTTPConnectionError

DOCKER_HUB_KATHARA_URL = "https://hub.docker.com/v2/repositories/kathara/?page_size=-1"

EXCLUDED_IMAGES = ['megalos-bgp-manager', 'katharanp']


class DockerHubApi(object):
    @staticmethod
    def get_images() -> filter:
        try:
            result = requests.get(DOCKER_HUB_KATHARA_URL)
        except requests.exceptions.ConnectionError as e:
            raise HTTPConnectionError(str(e))

        if result.status_code != 200:
            logging.debug("DockerHub replied with status code %s.", result.status_code)
            raise HTTPConnectionError("DockerHub replied with status code %s." % result.status_code)

        return filter(lambda x: not x['is_private'] and x['name'] not in EXCLUDED_IMAGES, result.json()['results'])



import argparse
import shutil
import sys
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class WipeCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara wipe',
            description=strings['wipe'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-f', '--force',
            required=False,
            action='store_true',
            help='Force the wipe.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-s', '--settings',
            required=False,
            action='store_true',
            help='Wipe the stored settings of the current user.'
        )

        group.add_argument(
            '-a', '--all',
            required=False,
            action='store_true',
            help='Wipe all Kathara devices and collision domains of all users. MUST BE ROOT FOR THIS OPTION.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        if not args['force']:
            utils.confirmation_prompt("Are you sure to wipe Kathara?", lambda: None, sys.exit)

        if args['settings']:
            Setting.wipe()
        else:
            if args['all'] and not utils.is_admin():
                raise Exception("You must be root in order to wipe all Kathara devices of all users.")

            Kathara.get_instance().wipe(all_users=bool(args['all']))

            vlab_dir = utils.get_vlab_temp_path(force_creation=False)
            shutil.rmtree(vlab_dir, ignore_errors=True)



import json
import logging
import os
from typing import Dict, Union, List

from deepdiff import DeepDiff

from .. import utils
from ..exceptions import MachineSignatureNotFoundError
from ..foundation.test.Test import Test
from ..model.Lab import Lab
from ..model.Machine import Machine


class BuiltInTest(Test):
    def __init__(self, lab: Lab) -> None:
        Test.__init__(self, lab)

    def create_signature(self) -> None:
        for (machine_name, machine) in self.lab.machines.items():
            logging.info("Building `builtin` signature for device %s..." % machine_name)

            machine_status = self._get_machine_status(machine)

            # Write the signature into the proper file
            with open("%s/%s.builtin" % (self.signature_path, machine_name), 'w') as machine_signature_file:
                machine_signature_file.write(json.dumps(machine_status, indent=4))

    def test(self) -> bool:
        test_passed = True

        for (machine_name, machine) in self.lab.machines.items():
            logging.info("Executing `builtin` tests for device %s..." % machine_name)

            machine_state = self._get_machine_status(machine)

            # Read the signature from machine file
            machine_signature_path = "%s/%s.builtin" % (self.signature_path, machine.name)
            if os.path.exists(machine_signature_path):
                with open(machine_signature_path, 'r') as machine_signature_file:
                    machine_signature = json.loads(machine_signature_file.read())
            else:
                raise MachineSignatureNotFoundError("Signature for device `%s` not found! Exiting..." % machine_name)

            # Save machine state into result file
            machine_result_path = "%s/%s.builtin" % (self.results_path, machine.name)
            with open(machine_result_path, 'w') as machine_result_file:
                machine_result_file.write(json.dumps(machine_state, indent=4))

            diff = self.check_signature(machine_signature, machine_state)
            test_passed = False if diff else test_passed

            machine_diff_path = "%s/%s.diff" % (self.results_path, machine.name)
            with open(machine_diff_path, 'w') as machine_diff_file:
                machine_diff_file.write(utils.format_headers("Builtin Test Result") + '\n')
                machine_diff_file.write(json.dumps(diff, indent=4) + "\n" if diff else "OK\n")
                machine_diff_file.write(utils.format_headers() + "\n\n")

        return test_passed

    def check_signature(self, signature: Union[Dict, str], status: Union[Dict, str]) -> Union[DeepDiff, List]:
        diff = DeepDiff(status, signature, ignore_order=True)

        if "iterable_item_added" in diff:
            diff["it_is"] = diff.pop("iterable_item_added")
        if "iterable_item_removed" in diff:
            diff["should_be"] = diff.pop("iterable_item_removed")

        return diff

    @staticmethod
    def _get_machine_status(machine: Machine) -> Dict:
        # Machine interfaces
        (ip_addr, _) = Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                                        machine_name=machine.name,
                                                        command="ip -j addr show"
                                                        )
        if ip_addr:
            ip_addr = json.loads(ip_addr)

            # Get only relevant information (interface name, state and list of address/prefix)
            ip_addr_clean = {}
            for info in ip_addr:
                ip_addr_clean[info['ifname']] = {'ip_addresses': [x["local"] + "/" + str(x["prefixlen"])
                                                                  for x in info["addr_info"]
                                                                  ],
                                                 'state': info['operstate']
                                                 }
        else:
            ip_addr_clean = {}

        # Machine routes
        (ip_route, _) = Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                                         machine_name=machine.name,
                                                         command="ip -j route show"
                                                         )

        ip_route = json.loads(ip_route) if ip_route else []

        # Machine opened ports
        (net_stat, _) = Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                                         machine_name=machine.name,
                                                         command="netstat -tuwln"
                                                         )
        # Remove Docker ports and header lines. Sort the array alphabetically.
        net_stat = sorted([filter(lambda x: "127.0.0.11" not in x, net_stat.splitlines())][2:]) if net_stat else []

        # Machine processes
        (processes, _) = Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                                          machine_name=machine.name,
                                                          command="ps -e -o command"
                                                          )
        # Remove header line and sort the array alphabetically.
        processes = sorted([x.strip() for x in processes.splitlines()[1:]]) if processes else []

        return {
            "interfaces": ip_addr_clean,
            "route": ip_route,
            "listening_ports": net_stat,
            "processes": processes
        }



import argparse
import logging
import sys
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class VstartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vstart',
            description=strings['vstart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=None,
            help='Start the device without opening a terminal window.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the device opening its terminal window.'
        )
        group.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the device in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        group.add_argument(
            '--num_terms',
            metavar='NUM_TERMS',
            required=False,
            help='Choose the number of terminals to open for the device.'
        )
        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be started.'
        )
        self.parser.add_argument(
            '--eth',
            dest='eths',
            metavar='N:CD',
            nargs='+',
            required=False,
            help='Set a specific interface on a collision domain.'
        )
        self.parser.add_argument(
            '-e', '--exec',
            required=False,
            dest='exec_commands',
            nargs='*',
            help='Execute a specific command in the device during startup.'
        )
        self.parser.add_argument(
            '--mem',
            required=False,
            help='Limit the amount of RAM available for this device.'
        )
        self.parser.add_argument(
            '--cpus',
            required=False,
            help='Limit the amount of CPU available for this device.'
        )
        self.parser.add_argument(
            '-i', '--image',
            required=False,
            help='Run this device with a specific Docker Image.'
        )
        self.parser.add_argument(
            '-H', '--no-hosthome',
            dest="no_hosthome",
            action="store_const",
            const=False,
            help='/hosthome dir will not be mounted inside the device.'
        )
        self.parser.add_argument(
            '--xterm',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        self.parser.add_argument(
            '--print',
            dest='dry_mode',
            required=False,
            action='store_true',
            help='Check if the device parameters are correct (dry run).'
        )
        self.parser.add_argument(
            '--bridged',
            required=False,
            action='store_true',
            help='Add a bridge interface to the device.'
        )
        self.parser.add_argument(
            '--port',
            dest='ports',
            metavar='[HOST:]GUEST[/PROTOCOL]',
            nargs='+',
            required=False,
            help='Map localhost port HOST to the internal port GUEST of the device for the specified PROTOCOL.'
        )
        self.parser.add_argument(
            '--sysctl',
            dest='sysctls',
            metavar='SYSCTL',
            nargs='+',
            required=False,
            help='Set sysctl option for the device.'
        )
        self.parser.add_argument(
            '--shell',
            required=False,
            help='Set the shell (sh, bash, etc.) that should be used inside the device.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        if args['dry_mode']:
            logging.info("Device configuration is correct. Exiting...")
            sys.exit(0)
        else:
            logging.info(utils.format_headers("Starting Device"))

        Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
            else Setting.get_instance().open_terminals
        Setting.get_instance().terminal = args['xterm'] or Setting.get_instance().terminal
        Setting.get_instance().device_shell = args['shell'] or Setting.get_instance().device_shell

        if args['privileged']:
            if not utils.is_admin():
                raise Exception("You must be root in order to start this Kathara device in privileged mode.")
            else:
                logging.warning("Running device with privileged capabilities, terminal won't open!")
                Setting.get_instance().open_terminals = False

        lab = Lab("kathara_vlab")
        lab.add_option('hosthome_mount', args['no_hosthome'])
        lab.add_option('shared_mount', False)
        lab.add_option('privileged_machines', args['privileged'])

        name = args.pop('name')

        device = lab.get_or_new_machine(name, **args)

        if args['eths']:
            for eth in args['eths']:
                try:
                    (iface_number, link_name) = eth.split(":")
                    lab.connect_machine_to_link(device.name, link_name, machine_iface_number=int(iface_number))
                except ValueError:
                    raise Exception("Interface number in `--eth %s` is not a number." % eth)

        lab.check_integrity()

        Kathara.get_instance().deploy_lab(lab)



from . import utils as setting_utils
from ....connectors.DockerHubApi import DockerHubApi
from ....exceptions import HTTPConnectionError
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....manager.Kathara import Kathara
from ....setting.Setting import Setting, DEFAULTS, AVAILABLE_DEBUG_LEVELS
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator
from ....utils import exec_by_platform
from ....validator.ImageValidator import ImageValidator
from ....validator.TerminalValidator import TerminalValidator

SHELLS_HINT = ["/bin/bash", "/bin/sh", "/bin/ash", "/bin/ksh", "/bin/zsh", "/bin/fish", "/bin/csh", "/bin/tcsh"]
TERMINALS_OSX = ["Terminal", "iTerm"]


class CommonOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # Manager Submenu
        managers = Kathara.get_available_managers_name()
        manager_type = Setting.get_instance().manager_type

        choose_manager_string = "Choose default manager"
        manager_menu = SelectionMenu(strings=[],
                                     title=choose_manager_string,
                                     subtitle=lambda: "Current: %s" % managers[manager_type],
                                     prologue_text="""Manager is the Engine used to run Kathara labs.
                                                   Default is `%s`.""" % managers[DEFAULTS['manager_type']],
                                     formatter=menu_formatter
                                     )

        for name, formatted_name in managers.items():
            manager_menu.append_item(FunctionItem(text=formatted_name,
                                                  function=self.update_manager_value,
                                                  args=[current_menu, name],
                                                  should_exit=True
                                                  )
                                     )

        manager_item = SubmenuItem(choose_manager_string, manager_menu, current_menu)

        # Image Selection Submenu
        image_string = "Choose default image"
        select_image_menu = SelectionMenu(strings=[],
                                          title=image_string,
                                          subtitle=setting_utils.current_string("image"),
                                          prologue_text="""Default Docker image when you start a lab or """
                                                        """a single Kathara device.
                                                        Default is `%s`.""" % DEFAULTS['image'],
                                          formatter=menu_formatter
                                          )

        try:
            for image in DockerHubApi.get_images():
                image_name = "%s/%s" % (image['namespace'], image['name'])

                select_image_menu.append_item(FunctionItem(text=image_name,
                                                           function=setting_utils.update_setting_value,
                                                           args=['image', image_name],
                                                           should_exit=True
                                                           )
                                              )
        except HTTPConnectionError:
            pass

        select_image_menu.append_item(FunctionItem(text="Choose another image",
                                                   function=setting_utils.read_value,
                                                   args=['image',
                                                         ImageValidator(),
                                                         'Write the name of a Docker image available on Docker Hub:',
                                                         'Docker Image not available neither on Docker Hub nor '
                                                         'in local repository!'
                                                         ],
                                                   should_exit=True
                                                   )
                                      )
        submenu_item = SubmenuItem(image_string, select_image_menu, current_menu)

        # Open Terminals Option
        open_terminals_string = "Automatically open terminals on startup"
        open_terminals_menu = SelectionMenu(strings=[],
                                            title=open_terminals_string,
                                            subtitle=setting_utils.current_bool("open_terminals"),
                                            prologue_text="""Determines if the device terminal should be opened """
                                                          """when starting it.
                                                          Default is %s."""
                                                          % setting_utils.format_bool(DEFAULTS['open_terminals']),
                                            formatter=menu_formatter
                                            )

        open_terminals_menu.append_item(FunctionItem(text="Yes",
                                                     function=setting_utils.update_setting_value,
                                                     args=['open_terminals', True],
                                                     should_exit=True
                                                     )
                                        )
        open_terminals_menu.append_item(FunctionItem(text="No",
                                                     function=setting_utils.update_setting_value,
                                                     args=['open_terminals', False],
                                                     should_exit=True
                                                     )
                                        )

        open_terminals_item = SubmenuItem(open_terminals_string, open_terminals_menu, current_menu)

        # Machine Shell Submenu
        machine_shell_string = "Choose device shell to be used"
        machine_shell_menu = SelectionMenu(strings=[],
                                           title=machine_shell_string,
                                           subtitle=setting_utils.current_string("device_shell"),
                                           formatter=menu_formatter,
                                           prologue_text="""The shell to use inside the device. 
                                                         **The application must be correctly installed in the Docker 
                                                         image used for the device!**
                                                         Default is `%s`, but it depends on the used Docker image.
                                                         """ % DEFAULTS['device_shell']
                                           )

        for shell in SHELLS_HINT:
            machine_shell_menu.append_item(FunctionItem(text=shell,
                                                        function=setting_utils.update_setting_value,
                                                        args=["device_shell", shell],
                                                        should_exit=True
                                                        )
                                           )
        machine_shell_menu.append_item(FunctionItem(text="Choose another shell",
                                                    function=setting_utils.read_value,
                                                    args=['device_shell',
                                                          RegexValidator(r"^(\w|/)+$"),
                                                          'Write the name of a shell:',
                                                          'Shell name is not valid!'
                                                          ],
                                                    should_exit=True
                                                    )
                                       )

        machine_shell_item = SubmenuItem(machine_shell_string, machine_shell_menu, current_menu)

        # Terminal Emulator Submenu
        # Linux Version
        def terminal_emulator_menu_linux():
            terminal_string = "Choose terminal emulator to be used"
            terminal_menu = SelectionMenu(strings=[],
                                          title=terminal_string,
                                          subtitle=setting_utils.current_string("terminal"),
                                          formatter=menu_formatter,
                                          prologue_text="""Terminal emulator application to be used for device """
                                                        """terminals. 
                                                        **The application must be correctly installed in """
                                                        """the host system!**
                                                        Default is `%s`.""" % DEFAULTS['terminal']
                                          )

            terminal_menu.append_item(FunctionItem(text="/usr/bin/xterm",
                                                   function=setting_utils.update_setting_value,
                                                   args=["terminal", "/usr/bin/xterm"],
                                                   should_exit=True
                                                   )
                                      )
            terminal_menu.append_item(FunctionItem(text="TMUX",
                                                   function=setting_utils.update_setting_value,
                                                   args=["terminal", "TMUX"],
                                                   should_exit=True
                                                   )
                                      )
            terminal_menu.append_item(FunctionItem(text="Choose another terminal emulator",
                                                   function=setting_utils.read_value,
                                                   args=['terminal',
                                                         TerminalValidator(),
                                                         'Write the path of a terminal emulator:',
                                                         'Terminal emulator is not valid! Install it before using it.'
                                                         ],
                                                   should_exit=True
                                                   )
                                      )

            return SubmenuItem(terminal_string, terminal_menu, current_menu)

        # macOS Version
        def terminal_emulator_menu_osx():
            terminal_string = "Choose terminal emulator to be used"
            terminal_menu = SelectionMenu(strings=[],
                                          title=terminal_string,
                                          subtitle=setting_utils.current_string("terminal"),
                                          formatter=menu_formatter,
                                          prologue_text="""Terminal emulator application to be used for device """
                                                        """terminals. 
                                                        **The application must be correctly installed in """
                                                        """the host system!**
                                                        Default is `Terminal`."""
                                          )

            for terminal in TERMINALS_OSX:
                terminal_menu.append_item(FunctionItem(text=terminal,
                                                       function=setting_utils.update_setting_value,
                                                       args=["terminal", terminal],
                                                       should_exit=True
                                                       )
                                          )
            terminal_menu.append_item(FunctionItem(text="TMUX",
                                                   function=setting_utils.update_setting_value,
                                                   args=["terminal", "TMUX"],
                                                   should_exit=True
                                                   )
                                      )

            return SubmenuItem(terminal_string, terminal_menu, current_menu)

        terminal_item = exec_by_platform(terminal_emulator_menu_linux, lambda: None, terminal_emulator_menu_osx)

        # Prefixes Submenu
        prefixes_string = "Choose Kathara prefixes"
        prefixes_menu = SelectionMenu(strings=[],
                                      title=prefixes_string,
                                      formatter=menu_formatter,
                                      prologue_text="""Prefixes assigned to the network and device names when deployed.
                                                    Default is `%s` and `%s`.""" % (DEFAULTS['net_prefix'],
                                                                                    DEFAULTS['device_prefix'])
                                      )

        net_prefix_item = FunctionItem(text=setting_utils.current_string("net_prefix",
                                                                         text="Insert Kathara networks prefix"),
                                       function=setting_utils.read_value,
                                       args=['net_prefix',
                                             RegexValidator(r"^[a-z]+_?[a-z_]+$"),
                                             'Write a Kathara networks prefix:',
                                             'Network Prefix must only contain lowercase letters and underscore.'
                                             ],
                                       should_exit=True
                                       )
        prefixes_menu.append_item(net_prefix_item)

        machine_prefix_item = FunctionItem(text=setting_utils.current_string("device_prefix",
                                                                             text="Insert Kathara devices prefix"),
                                           function=setting_utils.read_value,
                                           args=['device_prefix',
                                                 RegexValidator(r"^[a-z]+_?[a-z_]+$"),
                                                 'Write a Kathara devices prefix:',
                                                 'Device Prefix must only contain lowercase letters and underscore.'
                                                 ],
                                           should_exit=True
                                           )
        prefixes_menu.append_item(machine_prefix_item)

        prefixes_item = SubmenuItem(prefixes_string, prefixes_menu, current_menu)

        # Debug Level Submenu
        debug_level_string = "Choose logging level to be used"
        debug_level_menu = SelectionMenu(strings=[],
                                         title=debug_level_string,
                                         subtitle=setting_utils.current_string("debug_level"),
                                         formatter=menu_formatter,
                                         prologue_text="""Logging level of Kathara messages.
                                                       Default is `%s`.""" % DEFAULTS['debug_level']
                                         )

        for debug_level in AVAILABLE_DEBUG_LEVELS:
            debug_level_menu.append_item(FunctionItem(text=debug_level,
                                                      function=setting_utils.update_setting_value,
                                                      args=["debug_level", debug_level],
                                                      should_exit=True
                                                      )
                                         )

        debug_level_item = SubmenuItem(debug_level_string, debug_level_menu, current_menu)

        # Print Startup Logs Option
        print_startup_log_string = "Print Startup Logs on device startup"
        print_startup_log_menu = SelectionMenu(strings=[],
                                               title=open_terminals_string,
                                               subtitle=setting_utils.current_bool("print_startup_log"),
                                               formatter=menu_formatter,
                                               prologue_text="""When opening a device terminal, print its startup log.
                                                             Default is %s.""" %
                                                             setting_utils.format_bool(DEFAULTS['print_startup_log'])
                                               )

        print_startup_log_menu.append_item(FunctionItem(text="Yes",
                                                        function=setting_utils.update_setting_value,
                                                        args=['print_startup_log', True],
                                                        should_exit=True
                                                        )
                                           )
        print_startup_log_menu.append_item(FunctionItem(text="No",
                                                        function=setting_utils.update_setting_value,
                                                        args=['print_startup_log', False],
                                                        should_exit=True
                                                        )
                                           )

        print_startup_log_item = SubmenuItem(print_startup_log_string, print_startup_log_menu, current_menu)

        # Enable IPv6 Option
        enable_ipv6_string = "Enable IPv6"
        enable_ipv6_menu = SelectionMenu(strings=[],
                                         title=enable_ipv6_string,
                                         subtitle=setting_utils.current_bool("enable_ipv6"),
                                         formatter=menu_formatter,
                                         prologue_text="""This option enables IPv6 inside the devices.
                                                       Default is %s.""" %
                                                       setting_utils.format_bool(DEFAULTS['enable_ipv6']))

        enable_ipv6_menu.append_item(FunctionItem(text="Yes",
                                                  function=setting_utils.update_setting_value,
                                                  args=['enable_ipv6', True],
                                                  should_exit=True
                                                  )
                                     )
        enable_ipv6_menu.append_item(FunctionItem(text="No",
                                                  function=setting_utils.update_setting_value,
                                                  args=['enable_ipv6', False],
                                                  should_exit=True
                                                  )
                                     )

        enable_ipv6_item = SubmenuItem(enable_ipv6_string, enable_ipv6_menu, current_menu)

        current_menu.append_item(manager_item)
        current_menu.append_item(submenu_item)
        current_menu.append_item(open_terminals_item)
        current_menu.append_item(machine_shell_item)
        if terminal_item:
            current_menu.append_item(terminal_item)
        current_menu.append_item(prefixes_item)
        current_menu.append_item(debug_level_item)
        current_menu.append_item(print_startup_log_item)
        current_menu.append_item(enable_ipv6_item)

    def update_manager_value(self, current_menu: ConsoleMenu, value: str) -> None:
        setting_utils.update_setting_value("manager_type", value)

        # exit() does not work. So we force the menu to exit as the user selected the Exit option.
        # This is always the last item of the current menu.
        exit_option = len(current_menu.items) - 1
        current_menu.go_to(exit_option)

        clear_terminal()

        # Create a new menu (Manager could be changed so specific options need to be changed)
        new_menu = self.menu_factory.create_menu()
        new_menu.draw()
        new_menu.show()



from typing import Optional, Dict, Any

from ...foundation.setting.SettingsAddon import SettingsAddon

DEFAULTS = {
    "hosthome_mount": False,
    "shared_mount": True,
    "shared_cd": False,
    "remote_url": None,
    "cert_path": None
}


class DockerSettingsAddon(SettingsAddon):
    __slots__ = ['hosthome_mount', 'shared_mount', 'shared_cd', 'remote_url', 'cert_path']

    def __init__(self) -> None:
        self.hosthome_mount: bool = False
        self.shared_mount: bool = True
        self.shared_cd: bool = False
        self.remote_url: Optional[str] = None
        self.cert_path: Optional[str] = None

    def _to_dict(self) -> Dict[str, Any]:
        return {
            'hosthome_mount': self.hosthome_mount,
            'shared_mount': self.shared_mount,
            'shared_cd': self.shared_cd,
            'remote_url': self.remote_url,
            'cert_path': self.cert_path
        }



import subprocess

from . import ExternalItem


class CommandItem(ExternalItem):
    """
    A menu item to execute a console command
    """

    def __init__(self, text, command, arguments=None, menu=None, should_exit=False):
        """
        :ivar str command: The console command to be executed
        :ivar list[str] arguments: An optional list of string arguments to be passed to the command
        :ivar int exit_status: the exit status of the command, None if it hasn't been run yet
        """
        super(CommandItem, self).__init__(text=text, menu=menu, should_exit=should_exit)
        self.command = command

        if arguments:
            self.arguments = arguments
        else:
            self.arguments = []

        self.exit_status = None

    def action(self):
        """
        This class overrides this method
        """
        commandline = "{0} {1}".format(self.command, " ".join(self.arguments))
        try:
            completed_process = subprocess.run(commandline, shell=True)
            self.exit_status = completed_process.returncode
        except AttributeError:
            self.exit_status = subprocess.call(commandline, shell=True)

    def get_return(self):
        """
        :return: the exit status of the command
        :rtype: int
        """
        return self.exit_status



from terminaltables import SingleTable

strings = {
    "vstart": "Start a new Kathara device",
    "vclean": "Stop a single Kathara device",
    "vconfig": "Attach network interfaces to a running Kathara device",
    "lstart": "Start a Kathara lab",
    "lclean": "Stop a Kathara lab",
    "linfo": "Show information about a Kathara lab",
    "lrestart": "Restart a Kathara lab",
    "ltest": "Test a Kathara lab",
    "lconfig": "Attach network interfaces to a running Kathara device in a Kathara lab",
    "connect": "Connect to a Kathara device",
    "exec": "Execute a command in a Kathara device",
    "wipe": "Delete all Kathara devices and collision domains, optionally also delete settings",
    "list": "Show all running Kathara devices of the current user",
    "settings": "Show and edit Kathara settings",
    "check": "Check your system environment"
}

wiki_description = "For examples and further information visit: https://github.com/KatharaFramework/Kathara/wiki"


def formatted_strings() -> str:
    commands = []
    for item in strings.items():
        commands.append(list(item))

    commands_table = SingleTable(commands)
    commands_table.inner_heading_row_border = False
    commands_table.outer_border = False
    commands_table.inner_column_border = False

    return commands_table.table



from typing import Optional, Callable, Any, Tuple, List

from ....exceptions import SettingsError
from ....setting.Setting import Setting
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu import UserQuit
from ....trdparty.consolemenu.prompt_utils import InputResult
from ....trdparty.consolemenu.validators.base import BaseValidator

SAVED_STRING = "Saved successfully!\n"
PRESS_ENTER_STRING = "Press [Enter] to continue."

URL_REGEX = r'^(?:http)s?://'  # http:// or https://
URL_REGEX += r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
URL_REGEX += r'localhost|'  # localhost...
URL_REGEX += r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
URL_REGEX += r'(?::\d+)?$'  # optional port


def format_bool(value: bool) -> str:
    return "Yes" if value else "No"


def current_bool(attribute_name: str, text: Optional[str] = None) -> Callable[[], str]:
    return lambda: "%sCurrent: %s%s" % (text + " (" if text else "",
                                        (format_bool(getattr(Setting.get_instance(), attribute_name))),
                                        ")" if text else ""
                                        )


def current_string(attribute_name: str, text: Optional[str] = None) -> Callable[[], str]:
    return lambda: "%sCurrent: %s%s" % (text + " (" if text else "",
                                        getattr(Setting.get_instance(), attribute_name),
                                        ")" if text else ""
                                        )


def update_setting_value(attribute_name: str, value: Any, stdout: bool = True) -> None:
    reload = False

    if attribute_name == "manager_type" and Setting.get_instance().manager_type != value:
        reload = True

    setattr(Setting.get_instance(), attribute_name, value)

    try:
        if reload:
            Setting.get_instance().load_settings_addon()

        Setting.get_instance().check()

        Setting.get_instance().save()

        if stdout:
            print(SAVED_STRING)
    except SettingsError as e:
        print(str(e))

    if stdout:
        Screen().input(PRESS_ENTER_STRING)


def update_setting_values(attribute_values: List[Tuple[str, Any]]) -> None:
    for attribute_name, value in attribute_values:
        update_setting_value(attribute_name, value, stdout=False)

    print(SAVED_STRING)
    Screen().input(PRESS_ENTER_STRING)


def read_and_validate_value(prompt_msg: str, validator: BaseValidator, error_msg: str) -> InputResult:
    prompt_utils = PromptUtils(Screen())
    answer = prompt_utils.input(prompt=prompt_msg,
                                validators=validator,
                                enable_quit=True
                                )

    if not answer.validation_result:
        print(error_msg)

    return answer


def read_value(attribute_name: str, validator: BaseValidator, prompt_msg: str, error_msg: str) -> None:
    try:
        answer = read_and_validate_value(prompt_msg=prompt_msg,
                                         validator=validator,
                                         error_msg=error_msg
                                         )

        while not answer.validation_result:
            answer = read_and_validate_value(prompt_msg=prompt_msg,
                                             validator=validator,
                                             error_msg=error_msg
                                             )
    except UserQuit:
        return

    setattr(Setting.get_instance(), attribute_name, answer.input_string)
    Setting.get_instance().save()

    print(SAVED_STRING)

    Screen().input(PRESS_ENTER_STRING)
